from ovld import OvldBase, ovld


class Boop(OvldBase):
    @ovld(priority=100)
    def f(self, x: object):
        return self.f.next(x) * 3

    def f(self, x: int):
        return x * 2


print(Boop().f(7))



# from typing import TYPE_CHECKING
# from ovld import ovld
# from plum import dispatch as ovld
# from multimethod import multimethod as ovld
from runtype import multidispatch as ovld

# # if TYPE_CHECKING:
# from typing import overload


@ovld
def f(x: int) -> int:
    "A"
    return x * x


@ovld
def f(x: str) -> int:
    "COX"
    return int(x)


f
print(f(8))
print(f("8"))



from numbers import Number

import pytest

from .common import (
    fastcore_dispatch,
    multimethod_dispatch,
    multipledispatch_dispatch,
    ovld_dispatch,
    plum_dispatch,
    runtype_dispatch,
    singledispatch_dispatch,
)


class Animal:
    pass


class Mammal(Animal):
    pass


class Cat(Mammal):
    pass


class Dog(Mammal):
    pass


class Bird(Animal):
    pass


###############################
# multiple dispatch libraries #
###############################


def make_trivial(dispatch):
    @dispatch
    def trivial(x: Number):
        return "A"

    @dispatch
    def trivial(x: str):
        return "B"

    @dispatch
    def trivial(x: dict):
        return "C"

    @dispatch
    def trivial(x: list):
        return "D"

    @dispatch
    def trivial(x: Cat):
        return "E"

    @dispatch
    def trivial(x: Mammal):
        return "F"

    @dispatch
    def trivial(x: Animal):
        return "G"

    return trivial


##############
# isinstance #
##############


def trivial_isinstance(x):
    if isinstance(x, int | float):
        return "A"
    if isinstance(x, str):
        return "B"
    if isinstance(x, dict):
        return "C"
    if isinstance(x, list):
        return "D"
    if isinstance(x, Cat):
        return "E"
    if isinstance(x, Mammal):
        return "F"
    if isinstance(x, Animal):
        return "G"


####################
# Test definitions #
####################


def make_test(fn):
    @pytest.mark.benchmark(group="trivial")
    def test(benchmark):
        def run():
            return [
                fn(1),
                fn(3.5),
                fn("hello"),
                fn({}),
                fn([1, 2]),
                fn(Cat()),
                fn(Dog()),
                fn(Bird()),
            ]

        result = benchmark(run)
        assert result == list("AABCDEFG")

    return test


test_trivial_ovld = make_test(make_trivial(ovld_dispatch))
test_trivial_plum = make_test(make_trivial(plum_dispatch))
test_trivial_multimethod = make_test(make_trivial(multimethod_dispatch))
test_trivial_multipledispatch = make_test(
    make_trivial(multipledispatch_dispatch)
)
test_trivial_runtype = make_test(make_trivial(runtype_dispatch))
test_trivial_fastcore = make_test(make_trivial(fastcore_dispatch))
test_trivial_singledispatch = make_test(make_trivial(singledispatch_dispatch))

test_trivial_custom_isinstance = make_test(trivial_isinstance)



from timeit import timeit


def g1(x):
    return x * 7


def f1(xs):
    return [g1(x) for x in xs]


def mkf():
    def g2(x):
        return x * 7

    def f2(xs):
        return [g2(x) for x in xs]

    return f2


f2 = mkf()
print(timeit(lambda: f2(list(range(100))), number=100000))
print(timeit(lambda: f1(list(range(100))), number=100000))



from ovld import ovld, recurse
from ovld.core import OvldMC, extend_super


@ovld
def iterate_over_lists(xs: list):
    return [recurse(x) for x in xs]


@ovld
def iterate_over_dicts(xs: dict):
    return {k: recurse(v) for k, v in xs.items()}


@ovld(mixins=[iterate_over_lists, iterate_over_dicts])
def double(x):
    return x * 2


print(double([1, 2, 3]) == [2, 4, 6])
print(double({"x": 10, "y": 20}) == {"x": 20, "y": 40})


class IOL(metaclass=OvldMC):
    @ovld
    def __call__(self, xs: list):
        return [recurse(x) for x in xs]


class IOD(metaclass=OvldMC):
    @ovld
    def __call__(self, xs: dict):
        return {k: recurse(v) for k, v in xs.items()}


class Mul(IOL, IOD):
    def __init__(self, n):
        self.n = n

    @extend_super
    def __call__(self, x):
        return x * self.n


print(Mul(3)([1, 2, 3]))
print(Mul(3)({"x": 10, "y": 20}))



from numbers import Number
from typing import Literal

import pytest

from ovld.core import Ovld, OvldBase, ovld
from ovld.dependent import (
    Dependent,
    EndsWith,
    Equals,
    HasKeys,
    ParametrizedDependentType,
    Regexp,
    StartsWith,
    dependent_check,
)
from ovld.types import HasMethod
from ovld.utils import UsageError


class Bounded(ParametrizedDependentType):
    def default_bound(self, *parameters):
        return type(parameters[0])

    def check(self, value):
        min, max = self.parameters
        return min <= value <= max

    def __lt__(self, other):
        smin, smax = self.parameters
        omin, omax = other.parameters
        return (omin < smin and omax >= smax) or (omin <= smin and omax > smax)


def test_equality():
    assert Equals(0) == Equals(0)
    assert Equals[0] == Equals(0)
    assert Equals(0) != Equals(1)
    assert Bounded(0, 10) == Bounded(0, 10)


def test_dependent_type():
    @ovld
    def f(x: Equals(0)):
        return "zero"

    @f.register
    def f(x: Equals(1)):
        return "one"

    @f.register
    def f(x: int):
        return "nah"

    assert f(0) == "zero"
    assert f(1) == "one"
    assert f(2) == "nah"


def test_dependent_func():
    @ovld
    def f(x: Dependent[int, lambda x: x >= 0]):
        return "positive"

    @f.register
    def f(x: Dependent[int, lambda x: x < 0]):
        return "negative"

    assert f(0) == "positive"
    assert f(1) == "positive"
    assert f(2) == "positive"
    assert f(-2) == "negative"


def test_dependent_func():
    @dependent_check
    def InBetween(value: Number, min, max):
        return min <= value <= max

    @ovld
    def f(x: InBetween[1, 10]):
        return "A"

    @ovld
    def f(x: Number):
        return "B"

    @f.register
    def f(x):
        return "C"

    assert f(0) == "B"
    assert f(5) == "A"
    assert f("x") == "C"


def test_literal():
    @ovld
    def f(x: Literal[0]):
        return "zero"

    @f.register
    def f(x: Literal[1]):
        return "one"

    @f.register
    def f(x: Literal[2, 3]):
        return "two or three"

    @f.register
    def f(x: object):
        return "nah"

    assert f(0) == "zero"
    assert f(1) == "one"
    assert f(2) == "two or three"
    assert f(3) == "two or three"
    assert f(4) == "nah"


def test_many_literals():
    f = Ovld()

    n = 10

    for i in range(n):

        @ovld
        def f(x: Literal[i]):
            return x * x

    for i in range(n):
        assert f(i) == i * i

    with pytest.raises(TypeError, match="No method"):
        f(-1)

    with pytest.raises(TypeError, match="No method"):
        f(n)


def test_dependent_method():
    class Candy(OvldBase):
        def __init__(self, n):
            self.n = n

        @ovld
        def f(self, x: Equals(0)):
            return "zero"

        def f(self, x: Equals(1)):
            return "one"

        def f(self, x: int):
            return "nah"

    c = Candy(3)

    assert c.f(0) == "zero"
    assert c.f(1) == "one"
    assert c.f(2) == "nah"


def test_dependent_ambiguity():
    @ovld
    def f(s: Dependent[str, StartsWith("hell")]):
        return "A"

    @f.register
    def f(s: Dependent[str, EndsWith("ello")]):
        return "B"

    assert f("hell") == "A"
    with pytest.raises(TypeError, match="Ambiguous resolution"):
        f("hello")


def test_regexp():
    @ovld
    def f(s: Regexp[r"hell+o"]):
        return "hello"

    @ovld
    def f(s: str):
        return "bye"

    assert f("hello") == "hello"
    assert f("helllllllllo") == "hello"
    assert f("helllllllll") == "bye"
    assert f("oh helllllo there") == "hello"


def test_with_keys():
    @ovld
    def f(d: Dependent[dict, HasKeys("a")]):
        return "a"

    @f.register
    def f(d: Dependent[dict, HasKeys("b", "c")]):
        return "b|c"

    @f.register
    def f(d: dict):
        return "other"

    assert f({"a": 1}) == "a"
    assert f({"b": 2}) == "other"
    assert f({"b": 2, "c": 8}) == "b|c"
    with pytest.raises(TypeError):
        f({"a": 9, "b": 2, "c": 8})


def test_dependent_lists():
    HasLen = HasMethod["__len__"]

    @dependent_check
    def Nonempty(x: HasLen):
        return len(x) > 0

    @dependent_check
    def Length(x: HasLen, n):
        return len(x) == n

    @dependent_check
    def MinLength(x: HasLen, n):
        return len(x) >= n

    @ovld
    def f(li: Nonempty):
        return "nonempty"

    @f.register(priority=1)
    def f(li: Length[3]):
        return "three"

    @f.register(priority=1)
    def f(li: MinLength[5]):
        return ">=five"

    @f.register
    def f(li: HasLen):
        return "other"

    assert f([]) == "other"
    assert f([1]) == "nonempty"
    assert f([1, 2, 3]) == "three"
    assert f([1, 2, 3, 4]) == "nonempty"
    assert f([1, 2, 3, 4, 5]) == ">=five"
    assert f([1, 2, 3, 4, 5, 6, 7, 8]) == ">=five"


def test_bounded():
    @ovld
    def f(x: Dependent[Number, Bounded(0, 10)]):
        return "0-10"

    @f.register
    def f(x: Dependent[Number, Bounded(2, 6)]):
        return "2-6"

    assert Bounded(0, 10) > Bounded(2, 6)
    assert f(1) == "0-10"
    assert f(5) == "2-6"


def test_bound():
    @ovld
    def f(x: Dependent[int, lambda x: 1 / 0]):
        return 123

    @ovld
    def f(x):
        return x

    o = object()
    assert f(o) is o


def test_no_type_bound():
    class XXX(ParametrizedDependentType):
        pass

    with pytest.raises(UsageError):

        @ovld
        def f(x: XXX[1, 2, 3]):
            return "123"

        f(123)


def test_vs_catchall():
    @ovld
    def f(x: Dependent[Number, Bounded(0, 10)]):
        return "0-10"

    @f.register
    def f(x):
        return "other"

    assert f(1) == "0-10"
    assert f(25) == "other"
    assert f("zazz") == "other"


def test_or():
    o = Equals[0] | Equals[1]
    assert o.check(0)
    assert o.check(1)
    assert not o.check(2)

    @ovld
    def f(x: o):
        return 2

    assert f(0) == 2
    assert f(1) == 2


def test_and():
    a = Bounded[0, 100] & Bounded[-50, 50]
    assert not a.check(-50)
    assert a.check(1)
    assert not a.check(100)

    @ovld
    def f(x: a):
        return "yes"

    @ovld
    def f(x: object):
        return "no"

    assert f(-50) == "no"
    assert f(1) == "yes"
    assert f(100) == "no"


def test_or_and_errors():
    with pytest.raises(TypeError, match="unsupported operand"):
        Equals[0] | int
    with pytest.raises(TypeError, match="unsupported operand"):
        Equals[0] & int
    with pytest.raises(TypeError, match="unsupported operand"):
        int | Equals[0]
    with pytest.raises(TypeError, match="unsupported operand"):
        int & Equals[0]


def test_keyed_plus_other():
    # Tests a failure mode of generating dispatch code like HANDLER = HANDLERS[x]
    # if you also have to check a condition on y.

    for i in range(10):

        @ovld
        def f(x: Literal[i], y: Bounded[0, 100]):
            return "yes"

    @ovld
    def f(x: int, y: int):
        return "no"

    assert f(0, 50) == "yes"
    assert f(3, 50) == "yes"
    assert f(0, 150) == "no"



import inspect
import re
from dataclasses import dataclass
from itertools import count
from typing import TYPE_CHECKING, Any, Mapping, TypeVar, Union

from .types import (
    Intersection,
    Order,
    TypeRelationship,
    normalize_type,
    subclasscheck,
    typeorder,
)

_current = count()


@dataclass
class CodeGen:
    template: str
    substitutions: dict

    def mangle(self):
        renamings = {
            k: f"{{{k}__{next(_current)}}}" for k in self.substitutions
        }
        renamings["arg"] = "{arg}"
        new_subs = {
            newk[1:-1]: self.substitutions[k]
            for k, newk in renamings.items()
            if k in self.substitutions
        }
        return CodeGen(
            template=self.template.format(**renamings),
            substitutions=new_subs,
        )


def combine(master_template, args):
    fmts = []
    subs = {}
    for cg in args:
        mangled = cg.mangle()
        fmts.append(mangled.template)
        subs.update(mangled.substitutions)
    return CodeGen(master_template.format(*fmts), subs)


class DependentType:
    exclusive_type = False
    keyable_type = False

    def __init__(self, bound):
        self.bound = bound

    def __class_getitem__(cls, item):
        items = (item,) if not isinstance(item, tuple) else item
        return cls(*items)

    def with_bound(self, new_bound):  # pragma: no cover
        return type(self)(new_bound)

    def check(self, value):  # pragma: no cover
        raise NotImplementedError()

    def codegen(self):
        return CodeGen("{this}.check({arg})", {"this": self})

    def __typeorder__(self, other):
        if isinstance(other, DependentType):
            order = typeorder(self.bound, other.bound)
            if order is Order.SAME:
                # It isn't fully deterministic which of these will be called
                # because of set ordering between the types we compare
                if self < other:  # pragma: no cover
                    return TypeRelationship(Order.LESS, matches=False)
                elif other < self:  # pragma: no cover
                    return TypeRelationship(Order.MORE, matches=False)
                else:
                    return TypeRelationship(Order.NONE, matches=False)
            else:  # pragma: no cover
                return TypeRelationship(order, matches=False)
        elif (matches := subclasscheck(other, self.bound)) or subclasscheck(
            self.bound, other
        ):
            return TypeRelationship(Order.LESS, matches=matches)
        else:
            return TypeRelationship(Order.NONE, matches=False)

    def __lt__(self, other):
        return False

    def __or__(self, other):
        if not isinstance(other, DependentType):
            return NotImplemented
        return Or(self, other)

    def __and__(self, other):
        if not isinstance(other, DependentType):
            return NotImplemented
        return And(self, other)


class ParametrizedDependentType(DependentType):
    def __init__(self, *parameters, bound=None):
        super().__init__(
            self.default_bound(*parameters) if bound is None else bound
        )
        self.parameters = parameters

    @property
    def parameter(self):
        return self.parameters[0]

    def default_bound(self, *parameters):
        return None

    def with_bound(self, new_bound):
        return type(self)(*self.parameters, bound=new_bound)

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.parameters == other.parameters
            and self.bound == other.bound
        )

    def __hash__(self):
        return hash(self.parameters) ^ hash(self.bound)

    def __str__(self):
        params = ", ".join(map(repr, self.parameters))
        return f"{type(self).__name__}({params})"

    __repr__ = __str__


class FuncDependentType(ParametrizedDependentType):
    def default_bound(self, *_):
        fn = type(self).func
        return normalize_type(
            list(inspect.signature(fn).parameters.values())[0].annotation, fn
        )

    def __lt__(self, other):
        if len(self.parameters) != len(other.parameters):
            return False
        p1g = sum(
            p1 is Any and p2 is not Any
            for p1, p2 in zip(self.parameters, other.parameters)
        )
        p2g = sum(
            p2 is Any and p1 is not Any
            for p1, p2 in zip(self.parameters, other.parameters)
        )
        return p2g and not p1g

    def check(self, value):
        return type(self).func(value, *self.parameters)


def dependent_check(fn):
    t = type(fn.__name__, (FuncDependentType,), {"func": fn})
    if len(inspect.signature(fn).parameters) == 1:
        t = t()
    return t


class Equals(ParametrizedDependentType):
    keyable_type = True

    def default_bound(self, *parameters):
        return type(parameters[0])

    def check(self, value):
        return value in self.parameters

    @classmethod
    def keygen(cls):
        return "{arg}"

    def get_keys(self):
        return [self.parameter]

    def codegen(self):
        if len(self.parameters) == 1:
            return CodeGen("({arg} == {p})", {"p": self.parameter})
        else:
            return CodeGen("({arg} in {ps})", {"ps": self.parameters})


@dependent_check
def HasKeys(value: Mapping, *keys):
    return all(k in value for k in keys)


@dependent_check
def StartsWith(value: str, prefix):
    return value.startswith(prefix)


@dependent_check
def EndsWith(value: str, suffix):
    return value.endswith(suffix)


@dependent_check
def Regexp(value: str, regexp):
    return bool(re.search(pattern=regexp, string=value))


class Dependent:
    def __class_getitem__(cls, item):
        bound, dt = item
        if not isinstance(dt, DependentType):
            dt = dependent_check(dt)
        return dt.with_bound(bound)


class Or(DependentType):
    def __init__(self, *types, bound=None):
        self.types = types
        super().__init__(bound or self.default_bound())

    def default_bound(self):
        return Union[tuple([t.bound for t in self.types])]

    def codegen(self):
        template = " or ".join("{}" for t in self.types)
        return combine(template, [t.codegen() for t in self.types])

    def check(self, value):
        return any(t.check(value) for t in self.types)


class And(DependentType):
    def __init__(self, *types, bound=None):
        self.types = types
        super().__init__(bound or self.default_bound())

    def default_bound(self):
        bounds = frozenset(t.bound for t in self.types)
        return Intersection[tuple(bounds)]

    def codegen(self):
        template = " and ".join("{}" for t in self.types)
        return combine(template, [t.codegen() for t in self.types])

    def check(self, value):
        return all(t.check(value) for t in self.types)

    def __str__(self):
        return " & ".join(map(repr, self.types))

    __repr__ = __str__


if TYPE_CHECKING:  # pragma: no cover
    from typing import Annotated, TypeAlias

    T = TypeVar("T")
    A = TypeVar("A")
    Dependent: TypeAlias = Annotated[T, A]


__all__ = [
    "Dependent",
    "DependentType",
    "Equals",
    "HasKeys",
    "StartsWith",
    "EndsWith",
    "dependent_check",
]



"""Miscellaneous utilities."""

import functools


class Named:
    """A named object.

    This class can be used to construct objects with a name that will be used
    for the string representation.
    """

    def __init__(self, name):
        """Construct a named object.

        Arguments:
            name: The name of this object.
        """
        self.name = name

    def __repr__(self):
        """Return the object's name."""
        return self.name


BOOTSTRAP = Named("BOOTSTRAP")
MISSING = Named("MISSING")


def keyword_decorator(deco):
    """Wrap a decorator to optionally takes keyword arguments."""

    @functools.wraps(deco)
    def new_deco(fn=None, **kwargs):
        if fn is None:

            @functools.wraps(deco)
            def newer_deco(fn):
                return deco(fn, **kwargs)

            return newer_deco
        else:
            return deco(fn, **kwargs)

    return new_deco


class UsageError(Exception):
    pass


class Unusable:
    def __init__(self, message):
        self.__message = message

    def __call__(self, *args, **kwargs):
        raise UsageError(self.__message)

    def __getattr__(self, attr):
        raise UsageError(self.__message)


__all__ = [
    "BOOTSTRAP",
    "MISSING",
    "Named",
    "keyword_decorator",
]



from ovld import ovld, recurse
from ovld.dependent import Dependent, Equals, StartsWith

recurse.next()


@ovld
def test(x: Dependent[int, Equals(0)]):
    return "zero"


@ovld
def test(x: Dependent[int, Equals(1)]):
    return "one"


@ovld
def test(x: Dependent[str, StartsWith("hell")]):
    return "H"


@ovld
def test(x: Dependent[str, StartsWith("hello")]):
    return "H"


@ovld
def test(x: int):
    return "yes"


@ovld
def test(x: str):
    return "yes"


@ovld
def test(x: object):
    return "done"


@ovld(priority=10)
def test(x: object):
    return "start"


test.display_resolution(13)
test.display_resolution(0)
test.display_resolution(1)
test.display_resolution("hello")
test.map.display_methods()



import timeit

from multimethod import multimethod
from multipledispatch import dispatch
from plum import dispatch as plum

from ovld import ovld, recurse

# OVLD


@ovld
def smap(x: list, y: list):
    """One."""
    return [smap(a, b) for a, b in zip(x, y)]


@ovld
def smap(x: tuple, y: tuple):
    return tuple(smap(a, b) for a, b in zip(x, y))


@ovld
def smap(x: dict, y: dict):
    return {k: smap(v, y[k]) for k, v in x.items()}


@ovld
def smap(x: object, z: object):
    return x + z


# OVLD C


@ovld
def smap_c(x: list, y: list):
    """Two."""
    return [recurse(a, b) for a, b in zip(x, y)]


@ovld
def smap_c(x: tuple, y: tuple):
    return tuple(recurse(a, b) for a, b in zip(x, y))


@ovld
def smap_c(x: dict, y: dict):
    return {k: recurse(v, y[k]) for k, v in x.items()}


@ovld
def smap_c(x: object, y: object):
    return x + y


# multimethods


@multimethod
def smap_mm(x: list, y: list):
    """Three."""
    return [smap_mm(a, b) for a, b in zip(x, y)]


@multimethod
def smap_mm(x: tuple, y: tuple):
    return tuple(smap_mm(a, b) for a, b in zip(x, y))


@multimethod
def smap_mm(x: dict, y: dict):
    return {k: smap_mm(v, y[k]) for k, v in x.items()}


@multimethod
def smap_mm(x: object, y: object):
    return x + y


# multipledispatch


@dispatch(list, list)
def smap_md(x, y):
    """Four."""
    return [smap_md(a, b) for a, b in zip(x, y)]


@dispatch(tuple, tuple)
def smap_md(x, y):
    return tuple(smap_md(a, b) for a, b in zip(x, y))


@dispatch(dict, dict)
def smap_md(x, y):
    return {k: smap_md(v, y[k]) for k, v in x.items()}


@dispatch(object, object)
def smap_md(x, y):
    return x + y


# isinstance


def smap_ii(x, y):
    """Five."""
    if isinstance(x, dict) and isinstance(y, dict):
        return {k: smap_ii(v, y[k]) for k, v in x.items()}
    elif isinstance(x, tuple) and isinstance(y, tuple):
        return tuple(smap_ii(a, b) for a, b in zip(x, y))
    elif isinstance(x, list) and isinstance(y, list):
        return [smap_ii(a, b) for a, b in zip(x, y)]
    else:
        return x + y


# match statement


def smap_mh(x, y):
    """Six."""
    match (x, y):
        case ({}, {}):
            return {k: smap_mh(v, y[k]) for k, v in x.items()}
        case (tuple(), tuple()):
            return tuple(smap_mh(a, b) for a, b in zip(x, y))
        case ([*x], [*y]):
            return [smap_mh(a, b) for a, b in zip(x, y)]
        case _:
            return x + y


# plum


@plum
def smap_pl(x: list, y: list):
    """Three."""
    return [smap_pl(a, b) for a, b in zip(x, y)]


@plum
def smap_pl(x: tuple, y: tuple):
    return tuple(smap_pl(a, b) for a, b in zip(x, y))


@plum
def smap_pl(x: dict, y: dict):
    return {k: smap_pl(v, y[k]) for k, v in x.items()}


@plum
def smap_pl(x: object, y: object):
    return x + y


###


A = {"xs": list(range(50)), "ys": ("o", (6, 7))}
B = {"xs": list(range(10, 60)), "ys": ("x", (7, 6))}

results = {
    "smap": smap(A, B),
    "smap_mm": smap_mm(A, B),
    "smap_md": smap_md(A, B),
    "smap_pl": smap_pl(A, B),
    "smap_ii": smap_ii(A, B),
    "smap_mh": smap_mh(A, B),
    "smap_c": smap_c(A, B),
}

expected = results["smap"]

for k, v in results.items():
    assert v == expected, f"{k} failed"


# breakpoint()

print("smap_mm\t", 10 * timeit.timeit(lambda: smap_mm(A, B), number=10000))
print("smap_md\t", 10 * timeit.timeit(lambda: smap_md(A, B), number=10000))
print("smap_pl\t", 10 * timeit.timeit(lambda: smap_pl(A, B), number=10000))
print("smap_ov\t", 10 * timeit.timeit(lambda: smap(A, B), number=10000))
print("smap_c\t", 10 * timeit.timeit(lambda: smap_c(A, B), number=10000))
print("smap_ii\t", 10 * timeit.timeit(lambda: smap_ii(A, B), number=10000))
print("smap_mh\t", 10 * timeit.timeit(lambda: smap_mh(A, B), number=10000))



import inspect
import sys
import typing
from dataclasses import dataclass
from typing import Protocol, runtime_checkable

from ovld.utils import UsageError

from .mro import Order, TypeRelationship, subclasscheck, typeorder

try:
    from types import UnionType
except ImportError:  # pragma: no cover
    UnionType = None


def normalize_type(t, fn):
    from .dependent import DependentType, Equals

    if isinstance(t, str):
        t = eval(t, getattr(fn, "__globals__", {}))

    if t is type:
        t = type[object]
    elif t is typing.Any:
        t = object
    elif t is inspect._empty:
        t = object
    elif isinstance(t, typing._AnnotatedAlias):
        t = t.__origin__

    origin = getattr(t, "__origin__", None)
    if UnionType and isinstance(t, UnionType):
        return normalize_type(t.__args__, fn)
    elif origin is type:
        return t
    elif origin is typing.Union:
        return normalize_type(t.__args__, fn)
    elif origin is typing.Literal:
        return Equals(*t.__args__)
    elif origin and not getattr(t, "__args__", None):
        return t
    elif origin is not None:
        raise TypeError(
            f"ovld does not accept generic types except type, Union, Optional, Literal, but not: {t}"
        )
    elif isinstance(t, tuple):
        return typing.Union[tuple(normalize_type(t2, fn) for t2 in t)]
    elif isinstance(t, DependentType) and not t.bound:
        raise UsageError(
            f"Dependent type {t} has not been given a type bound. Please use Dependent[<bound>, {t}] instead."
        )
    else:
        return t


class MetaMC(type):
    def __new__(T, name, order):
        return super().__new__(T, name, (), {"order": order})

    def __init__(cls, name, order):
        pass

    def __typeorder__(cls, other):
        order = cls.order(other)
        if isinstance(order, bool):
            return NotImplemented
        else:
            return order

    def __subclasscheck__(cls, sub):
        result = cls.order(sub)
        if isinstance(result, TypeRelationship):
            return result.order in (Order.MORE, Order.SAME)
        else:
            return result


def class_check(condition):
    """Return a class with a subclassing relation defined by condition.

    For example, a dataclass is a subclass of `class_check(dataclasses.is_dataclass)`,
    and a class which name starts with "X" is a subclass of
    `class_check(lambda cls: cls.__name__.startswith("X"))`.

    Arguments:
        condition: A function that takes a class as an argument and returns
            True or False depending on whether it matches some condition.
    """
    return MetaMC(condition.__name__, condition)


def parametrized_class_check(fn):
    """Return a parametrized class checker.

    In essence, parametrized_class_check(fn)[X] will call fn(cls, X) in order
    to check whether cls matches the condition defined by fn and X.

    Arguments:
        fn: A function that takes a class and one or more additional arguments,
            and returns True or False depending on whether the class matches.
    """

    class _C:
        def __class_getitem__(_, arg):
            if not isinstance(arg, tuple):
                arg = (arg,)

            def arg_to_str(x):
                if isinstance(x, type):
                    return x.__name__
                else:
                    return repr(x)

            name = f"{fn.__name__}[{', '.join(map(arg_to_str, arg))}]"
            return MetaMC(name, lambda sub: fn(sub, *arg))

    _C.__name__ = fn.__name__
    _C.__qualname__ = fn.__qualname__
    return _C


def _getcls(ref):
    module, *parts = ref.split(".")
    curr = __import__(module)
    for part in parts:
        curr = getattr(curr, part)
    return curr


class Deferred:
    """Represent a class from an external module without importing it.

    For instance, `Deferred["numpy.ndarray"]` matches instances of
    numpy.ndarray, but it does not import numpy. When tested against a
    class, if the first part of class's `__module__` is `numpy`, then
    we do get the class and perform a normal issubclass check.

    If the module is already loaded, `Deferred` returns the class directly.

    Arguments:
        ref: A string starting with a module name representing the path
            to import a class.
    """

    def __class_getitem__(cls, ref):
        module, _ = ref.split(".", 1)
        if module in sys.modules:
            return _getcls(ref)

        def check(cls):
            full_cls_mod = getattr(cls, "__module__", None)
            cls_module = full_cls_mod.split(".", 1)[0] if full_cls_mod else None
            if cls_module == module:
                return issubclass(cls, _getcls(ref))
            else:
                return False

        return MetaMC(f"Deferred[{ref}]", check)


@parametrized_class_check
def Exactly(cls, base_cls):
    """Match the class but not its subclasses."""
    return TypeRelationship(
        order=Order.LESS if cls is base_cls else typeorder(base_cls, cls),
        matches=cls is base_cls,
    )


@parametrized_class_check
def StrictSubclass(cls, base_cls):
    """Match subclasses but not the base class."""
    return (
        isinstance(cls, type)
        and issubclass(cls, base_cls)
        and cls is not base_cls
    )


@parametrized_class_check
def Intersection(cls, *classes):
    """Match all classes."""
    matches = all(subclasscheck(cls, t) for t in classes)
    compare = [x for t in classes if (x := typeorder(t, cls)) is not Order.NONE]
    if not compare:
        return TypeRelationship(Order.NONE, matches=matches)
    elif any(x is Order.LESS or x is Order.SAME for x in compare):
        return TypeRelationship(Order.LESS, matches=matches)
    else:
        return TypeRelationship(Order.MORE, matches=matches)


@parametrized_class_check
def HasMethod(cls, method_name):
    """Match classes that have a specific method."""
    return hasattr(cls, method_name)


@runtime_checkable
@dataclass
class Dataclass(Protocol):
    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, "__dataclass_fields__") and hasattr(
            subclass, "__dataclass_params__"
        )


__all__ = [
    "Dataclass",
    "Deferred",
    "Exactly",
    "HasMethod",
    "Intersection",
    "StrictSubclass",
    "class_check",
    "parametrized_class_check",
]



from ovld import ovld, recurse

# def g(n):
#     @ovld
#     def f(x: int):
#         return x

#     @ovld
#     def f(x: int, y: int):
#         return recurse(x) + y

#     @ovld
#     def f(x: int, y: int, z: int):
#         return recurse(x, y) + z + n

#     return f


@ovld
def f(x: int):
    return x


@ovld
def f(x: int, y: int):
    return recurse(x) + y


@ovld
def f(x: int, y: int, z: object):
    return recurse(x, y) + z


# def flax(x: int, y: int, z: object, *, w: int = 3):
#     return x + y + z


print(f(3, z=7))



import timeit
from collections import defaultdict


class Foo(dict):
    pass


d1 = {"a": 1}
d2 = Foo({"a": 1})
d3 = defaultdict(lambda: 3)
d3["a"] = 1


def f1():
    for i in range(100):
        d1["a"]


def f2():
    for i in range(100):
        d2["a"]


def f3():
    for i in range(100):
        d3["a"]


def f4():
    for i in range(100):
        try:
            d1["a"]
        except KeyError:
            pass


print(timeit.timeit(f1, number=1000))
print(timeit.timeit(f2, number=1000))
print(timeit.timeit(f3, number=1000))
print(timeit.timeit(f4, number=1000))



import pytest

from ovld import call_next, recurse
from ovld.utils import UsageError


def test_unusable():
    with pytest.raises(
        UsageError, match="recurse.. can only be used from inside an @ovld"
    ):
        recurse()

    with pytest.raises(
        UsageError, match="call_next.. can only be used from inside an @ovld"
    ):
        call_next()

    with pytest.raises(
        UsageError, match="recurse.. can only be used from inside an @ovld"
    ):
        recurse.next



import sys
from functools import singledispatch

from fastcore.dispatch import typedispatch as fastcore_dispatch
from multimethod import multimethod as multimethod_dispatch
from multipledispatch import dispatch as _md_dispatch
from plum import dispatch as plum_dispatch
from runtype import multidispatch as runtype_dispatch

from ovld import ovld as ovld_dispatch


def _locate(fn):
    fr = sys._getframe(1)
    while fr and fn.__code__ not in fr.f_code.co_consts:
        fr = fr.f_back
    return fr.f_locals.get(fn.__name__, None)


def _getanns(fn):
    anns = fn.__annotations__.values()
    anns = [
        tuple(ann.__args__) if hasattr(ann, "__args__") else ann for ann in anns
    ]
    return anns


def multipledispatch_dispatch(fn):
    anns = _getanns(fn)
    existing = _locate(fn)
    if existing:
        existing.register(*anns)(fn)
        return existing
    else:
        return _md_dispatch(*anns)(fn)


def singledispatch_dispatch(fn):
    existing = _locate(fn)
    if existing:
        existing.register(fn)
        return existing
    else:
        return singledispatch(fn)


__all__ = [
    "multimethod_dispatch",
    "plum_dispatch",
    "runtype_dispatch",
    "ovld_dispatch",
    "multipledispatch_dispatch",
    "fastcore_dispatch",
    "singledispatch_dispatch",
]



from typing import Any

import torch
from torch import Tensor as T

from ovld import Dependent, ParametrizedDependentType, call_next, ovld


class Tensor(ParametrizedDependentType):
    def default_bound(self, *_):
        return T

    @property
    def shape(self):
        return [p for p in self.parameters if p is Any or isinstance(p, int)]

    @property
    def dtype(self):
        for p in self.parameters:
            if isinstance(p, torch.dtype):
                return p
        return Any

    def __lt__(self, other):
        if type(self) is not type(other):
            return False
        pairs = list(
            zip([self.dtype, *self.shape], [other.dtype, *other.shape])
        )
        return any(s1 is Any and s2 is not Any for s1, s2 in pairs) and not any(
            s2 for s1, s2 in pairs if s1 is not Any
        )

    def check(self, value):
        return (
            len(value.shape) == len(self.shape)
            and (self.dtype is Any or self.dtype is value.dtype)
            and all(
                s1 == s2
                for s1, s2 in zip(value.shape, self.shape)
                if s2 is not Any
            )
        )


def on_cpu(x):
    return x.device.type == "cpu"


@ovld(priority=1)
def convert(tensor: Dependent[T, on_cpu]):
    print("I'm on CPU")
    return call_next(tensor)


@ovld
def convert(t2d: Tensor[Any, Any]):
    return "A"


@ovld
def convert(t2d: Tensor[2, 2, torch.float64]):
    return "B"


@ovld
def convert(t2d: Tensor[2, 2]):
    return "C"


a = torch.zeros((2, 2))
print(convert(a))



import json
from collections import defaultdict
from pathlib import Path

file = Path(".benchmarks/Darwin-CPython-3.12-64bit/0011_report.json")
data = json.loads(file.read_text())


grouped = defaultdict(lambda: defaultdict(list))
for bench in data["benchmarks"]:
    match bench["name"].split("_"):
        case (_, _, category, variant):
            bench["category"] = category
            bench["variant"] = variant
        case (_, _, category):
            bench["category"] = category
            bench["variant"] = "normal"
    grouped[bench["group"]][category].append(bench)

columns = [
    ["custom", "custom", None],
    ["ovld", "ovld", "https://github.com/breuleux/ovld"],
    ["plum", "plum", "https://github.com/beartype/plum"],
    ["multimethod", "multim", "https://github.com/coady/multimethod"],
    [
        "multipledispatch",
        "multid",
        "https://github.com/mrocklin/multipledispatch/",
    ],
    ["runtype", "runtype", "https://github.com/erezsh/runtype"],
    ["fastcore", "fastcore", "https://github.com/fastai/fastcore"],
    [
        "singledispatch",
        "sd",
        "https://docs.python.org/3/library/functools.html#functools.singledispatch",
    ],
]
bench_names = ["trivial", "add", "multer", "ast", "calc", "fib", "tweaknum"]


template = "https://github.com/breuleux/ovld/tree/master/benchmarks/test_{}.py"

# print("| Benchmark | custom | ovld | md | plum | mm | runtype | sd |")
headers = [f"[{title}]({lnk})" if lnk else title for _, title, lnk in columns]

print(f"| Benchmark | {' | '.join(headers)} |")
print("| --- | ---: | ---: | ---: | ---: | ---: | ---: | ---: | ---: |")

for name in bench_names:
    group = grouped[name]
    best = min(
        min(bench["stats"]["median"] for bench in benches)
        for cat, benches in group.items()
        # if cat != "custom"
    )
    loc = template.format(name)
    line = [f"[{name}]({loc})"]
    for column, *_ in columns:
        benches = group.get(column, None)
        if not benches:
            line.append("x")
        else:
            value = min(bench["stats"]["median"] for bench in benches) / best
            line.append(f"{value:.2f}")

    print("|".join(["", *line, ""]))



from typing import Literal

import pytest

from .common import (
    multimethod_dispatch,
    ovld_dispatch,
    plum_dispatch,
)

###############################
# multiple dispatch libraries #
###############################


def make_fib(dispatch):
    @dispatch
    def fib(n: Literal[0]):
        return 0

    @dispatch
    def fib(n: Literal[1]):
        return 1

    @dispatch
    def fib(n: int):
        return fib(n - 1) + fib(n - 2)

    return fib


##########
# normal #
##########


def fib_normal(n):
    if n <= 1:
        return n
    else:
        return fib_normal(n - 1) + fib_normal(n - 2)


####################
# Test definitions #
####################


def make_test(fn):
    @pytest.mark.benchmark(group="fib")
    def test(benchmark):
        result = benchmark(fn, 8)
        assert result == 21

    return test


test_fib_ovld = make_test(make_fib(ovld_dispatch))
test_fib_plum = make_test(make_fib(plum_dispatch))
test_fib_multimethod = make_test(make_fib(multimethod_dispatch))

test_fib_custom = make_test(fib_normal)



from enum import Enum
from types import UnionType

from ovld import ovld
from ovld.core import OvldBase


class Poof[T]:
    pass


class Flax(OvldBase):
    @ovld
    def f(self, x: object, y: type[Poof[object]]):
        return 1234

    def f(self, frm: int, to: type[int]):
        return to(frm)

    def f(self, frm: float, to: type[float]):
        return to(frm)

    def f(self, frm: str, to: type[str]):
        return to(frm)

    def f(self, frm: str, to: type[Enum]):
        return to(frm)

    def f(self, frm: object, to: UnionType):
        for t in to.__args__:
            try:
                return self.deserialize(frm, t)
            except TypeError:
                continue
        else:
            return self.deserialize.next(frm, to)

    def f(self, frm: dict, to: type[Poof[object]]):
        model = self.model(to)
        if model is NotImplemented:
            return self.deserialize.next(frm, to)
        else:
            des = {
                k: self.deserialize(v, model.fields_by_name[k].type)
                for k, v in frm.items()
            }
            return model.builder(**des)

    def f(self, frm: dict, to: type[object]):
        model = self.model(to)
        if model is NotImplemented:
            return self.deserialize.next(frm, to)
        else:
            return self.deserialize(frm, model)

    # def f(self, frm: dict, typ: Model):
    #     des = {k: self.deserialize(v, typ.fields_by_name[k].type) for k, v in frm.items()}
    #     return typ.builder(**des)


@ovld
def f(x: object, y: type[Poof[object]]):
    return 12345


print(f(3, Poof[int]))
print(Flax().f(3, Poof[int]))


# print("=" * 80)
# print(compose_mro(list[int], {list[object]}, set()))

# print("=" * 80)
# print(compose_mro(Poof[int], {Poof[object]}, set()))



import json
import pickle
import sys
from pathlib import Path
from typing import Callable, Literal

import tomllib
import yaml
from jurigged.loop import loop

from ovld import call_next, ovld, recurse
from ovld.dependent import Dependent, ParametrizedDependentType

File = Dependent[Path, Path.is_file]


class FileType(ParametrizedDependentType):
    def default_bound(self, *_):
        return Path

    def check(self, f):
        return f.suffix == self.parameter


@ovld
def read(s: str):
    return recurse(Path(s))


@ovld
def read(f: File):
    return recurse(f.read_text(), format=f.suffix)


@ovld
def read(x: object, *, format: str | None = None, postprocess: Callable):
    if format is None:
        value = recurse(x)
    else:
        value = recurse(x, format=format)
    return postprocess(value)


@ovld
def read(channel: type(sys.stdin), *, format: str):
    return recurse(channel.read(), format=format)


@ovld
def read(f: File, *, format: Literal[".pickle"]):
    return pickle.load(f)


@ovld
def read(f: File, *, format: str):
    return recurse(f.read_text(), format=format)


@ovld
def read(s: str, *, format: Literal[".json"]):
    return json.loads(s)


@ovld
def read(s: str, *, format: Literal[".yaml"]):
    return yaml.safe_load(s)


@ovld
def read(s: str, *, format: Literal[".toml"]):
    return tomllib.loads(s)


@ovld
def read(xs: list):
    return [recurse(x) for x in xs]


@ovld(priority=-100)
def read(p: object, *, format: str = None):
    raise Exception(f"Could not read: {p}")


@read.variant(priority=100)
def read_only_here(f: File, *, format: str):
    if not f.absolute().is_relative_to("."):
        raise Exception("Can only read files in the current directory.")
    return call_next(f, format=format)


@loop(interface="rich")
def main():
    print(read(Path("../jurigged/pyproject.toml"), format=".toml"))
    print(read("world.yaml"))
    print(read(["world.yaml", "pyproject.toml"]))


if __name__ == "__main__":
    print(main)
    main()



import inspect
import math
import typing
from itertools import count
from types import CodeType

from .dependent import DependentType
from .mro import sort_types
from .recode import generate_dependent_dispatch
from .utils import MISSING


class GenericAliasMC(type):
    def __instancecheck__(cls, obj):
        return hasattr(obj, "__origin__")


class GenericAlias(metaclass=GenericAliasMC):
    pass


def is_type_of_type(t):
    return getattr(t, "__origin__", None) is type


class TypeMap(dict):
    """Represents a mapping from types to handlers.

    The mro of a type is considered when getting the handler, so setting the
    [object] key creates a default for all objects.

    typemap[some_type] returns a tuple of a handler and a "level" that
    represents the distance from the handler to the type `object`. Essentially,
    the level is the index of the type for which the handler was registered
    in the mro of `some_type`. So for example, `object` has level 0, a class
    that inherits directly from `object` has level 1, and so on.
    """

    def __init__(self):
        self.entries = {}
        self.types = set()

    def register(self, obj_t, handler):
        """Register a handler for the given object type."""
        self.clear()
        self.types.add(obj_t)
        s = self.entries.setdefault(obj_t, set())
        s.add(handler)

    def __missing__(self, obj_t):
        """Get the handler for the given type.

        The result is cached so that the normal dict getitem will find it
        the next time getitem is called.
        """
        results = {}
        groups = list(sort_types(obj_t, self.types))

        for lvl, grp in enumerate(reversed(groups)):
            for cls in grp:
                handlers = self.entries.get(cls, None)
                if handlers:
                    results.update({h: lvl for h in handlers})

        if results:
            self[obj_t] = results
            return results
        else:
            raise KeyError(obj_t)


class MultiTypeMap(dict):
    """Represents a mapping from tuples of types to handlers.

    The mro is taken into account to find a match. If multiple registered
    handlers match the tuple of types that's given, if one of the handlers is
    more specific than every other handler, that handler is returned.
    Otherwise, the resolution is considered ambiguous and an error is raised.

    Handler A, registered for types (A1, A2, ..., An), is more specific than
    handler B, registered for types (B1, B2, ..., Bn), if there exists n such
    that An is more specific than Bn, and for all n, either An == Bn or An is
    more specific than Bn. An is more specific than Bn if An is a direct or
    indirect subclass of Bn.

    In other words, [int, object] is more specific than [object, object] and
    less specific than [int, int], but it is neither less specific nor more
    specific than [object, int] (which means there is an ambiguity).
    """

    def __init__(self, name="_ovld", key_error=KeyError):
        self.maps = {}
        self.priorities = {}
        self.dependent = {}
        self.type_tuples = {}
        self.empty = MISSING
        self.key_error = key_error
        self.name = name
        self.dispatch_id = count()
        self.all = {}
        self.errors = {}

    def transform(self, obj):
        if isinstance(obj, GenericAlias):
            return type[obj]
        elif obj is typing.Any:
            return type[object]
        elif isinstance(obj, type):
            return type[obj]
        else:
            return type(obj)

    def mro(self, obj_t_tup):
        specificities = {}
        candidates = None
        nargs = len([t for t in obj_t_tup if not isinstance(t, tuple)])
        names = {t[0] for t in obj_t_tup if isinstance(t, tuple)}

        for i, cls in enumerate(obj_t_tup):
            if isinstance(cls, tuple):
                i, cls = cls

            try:
                results = self.maps[i][cls]
            except KeyError:
                results = {}

            results = {
                handler: spc
                for (handler, sig), spc in results.items()
                if sig.req_pos
                <= nargs
                <= (math.inf if sig.vararg else sig.max_pos)
                and not (sig.req_names - names)
            }

            try:
                vararg_results = self.maps[-1][cls]
            except KeyError:
                vararg_results = {}

            vararg_results = {
                handler: spc
                for (handler, sig), spc in vararg_results.items()
                if sig.req_pos <= nargs and i >= sig.max_pos
            }

            results.update(vararg_results)

            if candidates is None:
                candidates = set(results.keys())
            else:
                candidates &= results.keys()
            for c in candidates:
                specificities.setdefault(c, []).append(results[c])

        candidates = [
            (c, self.priorities.get(c, 0), tuple(specificities[c]))
            for c in candidates
        ]

        # The sort ensures that if candidate A dominates candidate B, A will
        # appear before B in the list. That's because it must dominate all
        # other possibilities on all arguments, so the sum of all specificities
        # has to be greater.
        # Note: priority is always more important than specificity
        candidates.sort(key=lambda cspc: (cspc[1], sum(cspc[2])), reverse=True)

        self.all[obj_t_tup] = {
            getattr(c[0], "__code__", None) for c in candidates
        }

        processed = set()

        def _pull(candidates):
            candidates = [
                (c, a, b) for (c, a, b) in candidates if c not in processed
            ]
            if not candidates:
                return
            rval = [candidates[0]]
            c1, p1, spc1 = candidates[0]
            for c2, p2, spc2 in candidates[1:]:
                if p1 > p2 or (
                    spc1 != spc2 and all(s1 >= s2 for s1, s2 in zip(spc1, spc2))
                ):
                    # Candidate 1 dominates candidate 2
                    continue
                else:
                    processed.add(c2)
                    # Candidate 1 does not dominate candidate 2, so we add it
                    # to the list.
                    rval.append((c2, p2, spc2))
            yield rval
            if len(rval) >= 1:
                yield from _pull(candidates[1:])

        return list(_pull(candidates))

    def register(self, sig, handler):
        """Register a handler for a tuple of argument types.

        Arguments:
            sig: A Signature object.
            handler: A function to handle the tuple.
        """
        self.clear()

        obj_t_tup = sig.types
        entry = (handler, sig)
        if not obj_t_tup:
            self.empty = entry

        self.priorities[handler] = sig.priority
        self.type_tuples[handler] = obj_t_tup
        self.dependent[handler] = any(
            isinstance(t[1] if isinstance(t, tuple) else t, DependentType)
            for t in obj_t_tup
        )

        for i, cls in enumerate(obj_t_tup):
            if isinstance(cls, tuple):
                i, cls = cls
            if i not in self.maps:
                self.maps[i] = TypeMap()
            self.maps[i].register(cls, entry)

        if sig.vararg:  # pragma: no cover
            # TODO: either add this back in, or remove it
            if -1 not in self.maps:
                self.maps[-1] = TypeMap()
            self.maps[-1].register(object, entry)

    def display_methods(self):
        for h, prio in sorted(self.priorities.items(), key=lambda kv: -kv[1]):
            prio = f"[{prio}]"
            width = 6
            print(f"{prio:{width}} \033[1m{h.__name__}\033[0m")
            co = h.__code__
            print(f"{'':{width-2}} @ {co.co_filename}:{co.co_firstlineno}")

    def display_resolution(self, *args, **kwargs):
        def dependent_match(tup, args):
            for t, a in zip(tup, args):
                if isinstance(t, tuple):
                    t = t[1]
                    a = a[1]
                if isinstance(t, DependentType) and not t.check(a):
                    return False
            return True

        message = "No method will be called."
        argt = [
            *map(self.transform, args),
            *[(k, self.transform(v)) for k, v in kwargs.items()],
        ]
        finished = False
        rank = 1
        for grp in self.mro(tuple(argt)):
            grp.sort(key=lambda x: x[0].__name__)
            match = [
                dependent_match(
                    self.type_tuples[handler], [*args, *kwargs.items()]
                )
                for handler, _, _ in grp
            ]
            ambiguous = len([m for m in match if m]) > 1
            for m, (handler, prio, spec) in zip(match, grp):
                color = "\033[0m"
                if finished:
                    bullet = "--"
                    color = "\033[1;90m"
                elif not m:
                    bullet = "!="
                    color = "\033[1;90m"
                elif ambiguous:
                    bullet = "=="
                    color = "\033[1;31m"
                else:
                    bullet = f"#{rank}"
                    if rank == 1:
                        message = f"{handler.__name__} will be called first."
                        color = "\033[1;32m"
                    rank += 1
                spec = ".".join(map(str, spec))
                lvl = f"[{prio}:{spec}]"
                width = 2 * len(args) + 6
                print(f"{color}{bullet} {lvl:{width}} {handler.__name__}")
                co = handler.__code__
                print(
                    f"   {'':{width-1}}@ {co.co_filename}:{co.co_firstlineno}\033[0m"
                )
            if ambiguous:
                message += " There is ambiguity between multiple matching methods, marked '=='."
                finished = True
        print("Resolution:", message)

    def wrap_dependent(self, tup, handlers, group, next_call):
        handlers = list(handlers)
        htup = [(h, self.type_tuples[h]) for h in handlers]
        slf = (
            "self, "
            if inspect.getfullargspec(handlers[0]).args[0] == "self"
            else ""
        )
        return generate_dependent_dispatch(
            tup,
            htup,
            next_call,
            slf,
            name=f"{self.name}.specialized_dispatch_{next(self.dispatch_id)}",
            err=self.key_error(tup, group),
            nerr=self.key_error(tup, ()),
        )

    def resolve(self, obj_t_tup):
        results = self.mro(obj_t_tup)
        if not results:
            raise self.key_error(obj_t_tup, ())

        funcs = []
        for group in reversed(results):
            handlers = [fn for (fn, _, _) in group]
            dependent = any(self.dependent[fn] for (fn, _, _) in group)
            if dependent:
                nxt = self.wrap_dependent(
                    obj_t_tup, handlers, group, funcs[-1] if funcs else None
                )
            elif len(group) != 1:
                nxt = None
            else:
                nxt = handlers[0]
            codes = [h.__code__ for h in handlers if hasattr(h, "__code__")]
            funcs.append((nxt, codes))

        funcs.reverse()

        parents = []
        for group, (func, codes) in zip(results, funcs):
            tups = (
                [obj_t_tup]
                if not parents
                else [(parent, *obj_t_tup) for parent in parents]
            )
            if func is None:
                for tup in tups:
                    self.errors[tup] = self.key_error(obj_t_tup, group)
                break
            else:
                for tup in tups:
                    self[tup] = func
            if not codes:
                break
            parents = codes

        return True

    def __missing__(self, obj_t_tup):
        if obj_t_tup and isinstance(obj_t_tup[0], CodeType):
            real_tup = obj_t_tup[1:]
            self[real_tup]
            if obj_t_tup[0] not in self.all[real_tup]:
                return self[real_tup]
            elif obj_t_tup in self.errors:
                raise self.errors[obj_t_tup]
            elif obj_t_tup in self:  # pragma: no cover
                # PROBABLY not reachable
                return self[obj_t_tup]
            else:
                raise self.key_error(real_tup, ())

        if not obj_t_tup:
            if self.empty is MISSING:  # pragma: no cover
                # Might not be reachable because of codegen
                raise self.key_error(obj_t_tup, ())
            else:
                return self.empty[0]

        self.resolve(obj_t_tup)
        if obj_t_tup in self.errors:
            raise self.errors[obj_t_tup]
        else:
            return self[obj_t_tup]



from timeit import timeit

from plum import dispatch

from ovld import ovld as dispatch


def f1(x):
    match x:
        case str():
            return "string"
        case int():
            return "integer"


# from functools import singledispatch as dispatch


@dispatch
def f2(x: str):
    return "string"


@dispatch
def f2(x: int):
    return "integer"


test = [
    1,
    7,
    "5",
    "bob",
    1564813548,
    "Four score and seven years ago",
    456,
    "879",
    "I can't understand why my fingers messed that up",
    852,
    1234567890,
    "More of the same",
    "Spam spam spam",
    1,
    1,
    2,
    3,
    5,
    8,
    13,
    21,
    33,
    54,
    87,
    "Fizz",
    "Buzz",
]


def run(f, test):
    for one in test:
        f(one)


print(timeit(lambda: run(f1, test), number=10000))
print(timeit(lambda: run(f2, test), number=10000))



# # import dis

# # z = 10


# # def foo(x):
# #     def faf(y):
# #         return x + y + z

# #     return faf


# # fn = foo(4)

# # bytecode = dis.Bytecode(fn)
# # print("=" * 80)
# # for instr in bytecode:
# #     print(instr)

# # breakpoint()


# from wrapt import ObjectProxy

# from ovld import ovld, recurse


# class AnnotatedProxy(ObjectProxy):
#     def __init__(self, wrapped, **ann):
#         super().__init__(wrapped)
#         self._self_ann = ann

#     @property
#     def _(self):
#         return self._self_ann


# # @ovld(priority=10)
# # def f(x: object):
# #     return recurse.next(x)


# @ovld
# def f(xs: list):
#     return [recurse(x) for x in xs]


# @ovld
# def f(xs: dict):
#     return {recurse(k): recurse(v) for k, v in xs.items()}


# @ovld
# def f(x: str):
#     return x


# @ovld
# def f(x: int):
#     return x / 0


# # # f({"a": [[[[1, 2, 3]]]]})
# # # breakpoint()
# # res = f({"a": [[[[1, 2, 3]]]]})
# # print(res)

# # data = {"a": 2}
# # print(f(AnnotatedProxy(data, x=4)))


from ovld import call_next, ovld, recurse


@ovld(priority=10)
def f(x: object):
    return call_next(x)


@ovld
def f(xs: list):
    return [recurse(x) for x in xs]


@ovld
def f(xs: dict):
    return {recurse(k): recurse(v) for k, v in xs.items()}


@ovld
def f(x: str):
    return x


@ovld
def f(x: int):
    return x / 0


print(f([[1, 2]]))



from ovld import ovld


def foo(dispatch, T, fallback):
    @dispatch
    def f(x: T):
        return x * 2

    @dispatch
    def f(x: object):
        return fallback

    return f


f1 = foo(ovld, int, "OK")
print(f1(1))
print(f1("x"))
print(f1(object()))

f2 = foo(ovld, str, "NONONO")
print(f1(1))
print(f1("x"))
print(f1(object()))



from ovld import ovld

# @ovld
# def fact(n: Literal[0]):
#     return 1
# @ovld
# def fact(n: Dependent[int, lambda n: n > 0]):
#     return n * recurse(n - 1)
# print(fact(5))
# fact(-1)   # Error!
# @ovld
# def add(x: list, y: list):
#     return [recurse(a, b) for a, b in zip(x, y)]
# @ovld
# def add(x: tuple, y: tuple):
#     return tuple(recurse(a, b) for a, b in zip(x, y))
# @ovld
# def add(x: dict, y: dict):
#     return {k: recurse(v, y[k]) for k, v in x.items()}
# @ovld
# def add(x: object, y: object):
#     return x + y
# @add.variant
# def bad(self, x: object, y: object):
#     raise Exception("Bad.")
# add([[[1]]], [[[[2]]]])
# @ovld
# def f(cls: type[list[object]], xs: list):
#     return [recurse(cls.__args__[0], x) for x in xs]
# @ovld
# def f(cls: type[int], x: int):
#     return x * 2
# assert f(list[int], [1, 2, 3]) == [2, 4, 6]
# f(list[int], [1, "X", 3])  # type error
from ovld.types import Exactly, StrictSubclass


@ovld
def f(x: Exactly[BaseException]):
    return "=BaseException"


@ovld
def f(x: StrictSubclass[Exception]):
    return ">Exception"


assert f(TypeError()) == ">Exception"
assert f(BaseException()) == "=BaseException"

f(Exception())  # ERROR!



import ast
import inspect
import linecache
import textwrap
from ast import _splitlines_no_ff as splitlines
from functools import reduce
from itertools import count
from types import CodeType, FunctionType

from .dependent import DependentType
from .utils import Unusable, UsageError

recurse = Unusable(
    "recurse() can only be used from inside an @ovld-registered function."
)
call_next = Unusable(
    "call_next() can only be used from inside an @ovld-registered function."
)


dispatch_template = """
from ovld.utils import MISSING

def __DISPATCH__(self, {args}):
    {inits}
    {body}
    {call}
"""


call_template = """
method = self.map[({lookup})]
return method({posargs})
"""


def instantiate_code(symbol, code, inject={}):
    virtual_file = f"<ovld{hash(code)}>"
    linecache.cache[virtual_file] = (None, None, splitlines(code), virtual_file)
    code = compile(source=code, filename=virtual_file, mode="exec")
    glb = {**inject}
    exec(code, glb, glb)
    return glb[symbol]


# # Previous version: generate a temporary file
# def instantiate_code(symbol, code, inject={}):
#     tf = tempfile.NamedTemporaryFile("w")
#     _tempfiles.append(tf)
#     tf.write(code)
#     tf.flush()
#     glb = runpy.run_path(tf.name)
#     rval = glb[symbol]
#     rval.__globals__.update(inject)
#     return rval


def generate_dispatch(arganal):
    def join(li, sep=", ", trail=False):
        li = [x for x in li if x]
        rval = sep.join(li)
        if len(li) == 1 and trail:
            rval += ","
        return rval

    spr, spo, pr, po, kr, ko = arganal.compile()

    inits = set()

    kwargsstar = ""
    targsstar = ""

    args = []
    body = [""]
    posargs = ["self.obj" if arganal.is_method else ""]
    lookup = []

    i = 0

    for name in spr + spo:
        if name in spr:
            args.append(name)
        else:
            args.append(f"{name}=MISSING")
        posargs.append(name)
        lookup.append(f"{arganal.lookup_for(i)}({name})")
        i += 1

    if len(po) <= 1:
        # If there are more than one non-strictly positional optional arguments,
        # then all positional arguments are strictly positional, because if e.g.
        # x and y are optional we want x==MISSING to imply that y==MISSING, but
        # that only works if y cannot be provided as a keyword argument.
        args.append("/")

    for name in pr + po:
        if name in pr:
            args.append(name)
        else:
            args.append(f"{name}=MISSING")
        posargs.append(name)
        lookup.append(f"{arganal.lookup_for(i)}({name})")
        i += 1

    if len(po) > 1:
        args.append("/")

    if kr or ko:
        args.append("*")

    for name in kr:
        lookup_fn = (
            "self.map.transform"
            if name in arganal.complex_transforms
            else "type"
        )
        args.append(f"{name}")
        posargs.append(f"{name}={name}")
        lookup.append(f"({name!r}, {lookup_fn}({name}))")

    for name in ko:
        args.append(f"{name}=MISSING")
        kwargsstar = "**KWARGS"
        targsstar = "*TARGS"
        inits.add("KWARGS = {}")
        inits.add("TARGS = []")
        body.append(f"if {name} is not MISSING:")
        body.append(f"    KWARGS[{name!r}] = {name}")
        body.append(
            f"    TARGS.append(({name!r}, {arganal.lookup_for(name)}({name})))"
        )

    posargs.append(kwargsstar)
    lookup.append(targsstar)

    fullcall = call_template.format(
        lookup=join(lookup, trail=True),
        posargs=join(posargs),
    )

    calls = []
    if spo or po:
        req = len(spr + pr)
        for i, arg in enumerate(spo + po):
            call = call_template.format(
                lookup=join(lookup[: req + i], trail=True),
                posargs=join(posargs[: req + i + 1]),
            )
            call = textwrap.indent(call, "    ")
            calls.append(f"\nif {arg} is MISSING:{call}")
    calls.append(fullcall)

    code = dispatch_template.format(
        inits=join(inits, sep="\n    "),
        args=join(args),
        body=join(body, sep="\n    "),
        call=textwrap.indent("".join(calls), "    "),
    )
    return instantiate_code("__DISPATCH__", code)


class GenSym:
    def __init__(self, prefix):
        self.prefix = prefix
        self.count = count()
        self.variables = {}

    def add(self, value):
        if isinstance(value, (int, float, str)):
            return repr(value)
        id = f"{self.prefix}{next(self.count)}"
        self.variables[id] = value
        return id


def generate_dependent_dispatch(tup, handlers, next_call, slf, name, err, nerr):
    def to_dict(tup):
        return dict(
            entry if isinstance(entry, tuple) else (i, entry)
            for i, entry in enumerate(tup)
        )

    def argname(x):
        return f"ARG{x}" if isinstance(x, int) else x

    def argprovide(x):
        return f"ARG{x}" if isinstance(x, int) else f"{x}={x}"

    def codegen(typ, arg):
        cg = typ.codegen()
        return cg.template.format(
            arg=arg, **{k: gen.add(v) for k, v in cg.substitutions.items()}
        )

    tup = to_dict(tup)
    handlers = [(h, to_dict(types)) for h, types in handlers]
    gen = GenSym(prefix="INJECT")
    conjs = []

    exclusive = False
    keyexpr = None
    keyed = None
    for k in tup:
        featured = set(types[k] for h, types in handlers)
        if len(featured) == len(handlers):
            possibilities = set(type(t) for t in featured)
            focus = possibilities.pop()
            # Possibilities is now empty if only one type of DependentType

            if not possibilities:
                if getattr(focus, "keyable_type", False):
                    all_keys = [
                        {key: h for key in types[k].get_keys()}
                        for h, types in handlers
                    ]
                    keyed = reduce(lambda a, b: {**a, **b}, all_keys)
                    if (
                        len(keyed) == sum(map(len, all_keys))
                        and len(featured) < 4
                    ):
                        exclusive = True
                        keyexpr = None
                    else:
                        keyexpr = focus.keygen().format(arg=argname(k))

                else:
                    exclusive = getattr(focus, "exclusive_type", False)

    for i, (h, types) in enumerate(handlers):
        relevant = [k for k in tup if isinstance(types[k], DependentType)]
        if len(relevant) > 1:
            # The keyexpr method only works if there is only one condition to check.
            keyexpr = keyed = None
        codes = [codegen(types[k], argname(k)) for k in relevant]
        conj = " and ".join(codes)
        if not conj:  # pragma: no cover
            # Not sure if this can happen
            conj = "True"
        conjs.append(conj)

    argspec = ", ".join(argname(x) for x in tup)
    argcall = ", ".join(argprovide(x) for x in tup)

    body = []
    if keyexpr:
        body.append(f"HANDLER = {gen.add(keyed)}.get({keyexpr}, FALLTHROUGH)")
        body.append(f"return HANDLER({slf}{argcall})")

    elif exclusive:
        for i, conj in enumerate(conjs):
            body.append(f"if {conj}: return HANDLER{i}({slf}{argcall})")
        body.append(f"return FALLTHROUGH({slf}{argcall})")

    else:
        for i, conj in enumerate(conjs):
            body.append(f"MATCH{i} = {conj}")

        summation = " + ".join(f"MATCH{i}" for i in range(len(handlers)))
        body.append(f"SUMMATION = {summation}")
        body.append("if SUMMATION == 1:")
        for i, (h, types) in enumerate(handlers):
            body.append(f"    if MATCH{i}: return HANDLER{i}({slf}{argcall})")
        body.append("elif SUMMATION == 0:")
        body.append(f"    return FALLTHROUGH({slf}{argcall})")
        body.append("else:")
        body.append(f"    raise {gen.add(err)}")

    body_text = textwrap.indent("\n".join(body), "    ")
    code = f"def __DEPENDENT_DISPATCH__({slf}{argspec}):\n{body_text}"

    inject = gen.variables
    for i, (h, types) in enumerate(handlers):
        inject[f"HANDLER{i}"] = h

    def raise_error(*args, **kwargs):
        raise nerr

    inject["FALLTHROUGH"] = (next_call and next_call[0]) or raise_error

    fn = instantiate_code(
        symbol="__DEPENDENT_DISPATCH__", code=code, inject=inject
    )
    return rename_function(fn, name)


_current = count()


class Conformer:
    __slots__ = ("code", "orig_fn", "renamed_fn", "ovld")

    def __init__(self, ovld, orig_fn, renamed_fn):
        self.ovld = ovld
        self.orig_fn = orig_fn
        self.renamed_fn = renamed_fn
        self.code = orig_fn.__code__

    def __conform__(self, new):
        if isinstance(new, FunctionType):
            new_fn = new
            new_code = new.__code__
        else:
            new_fn = None
            new_code = new

        self.ovld.unregister(self.orig_fn)

        if new_fn is None:  # pragma: no cover
            if new_code is None:
                return
            ofn = self.orig_fn
            new_fn = FunctionType(
                new_code,
                ofn.__globals__,
                ofn.__name__,
                ofn.__defaults__,
                ofn.__closure__,
            )
            new_fn.__annotations__ = ofn.__annotations__

        self.ovld.register(new_fn)

        from codefind import code_registry

        code_registry.update_cache_entry(self, self.code, new_code)

        self.code = new_code


def rename_code(co, newname):  # pragma: no cover
    if hasattr(co, "replace"):
        if hasattr(co, "co_qualname"):
            return co.replace(co_name=newname, co_qualname=newname)
        else:
            return co.replace(co_name=newname)
    else:
        return type(co)(
            co.co_argcount,
            co.co_kwonlyargcount,
            co.co_nlocals,
            co.co_stacksize,
            co.co_flags,
            co.co_code,
            co.co_consts,
            co.co_names,
            co.co_varnames,
            co.co_filename,
            newname,
            co.co_firstlineno,
            co.co_lnotab,
            co.co_freevars,
            co.co_cellvars,
        )


def rename_function(fn, newname):
    """Create a copy of the function with a different name."""
    newcode = rename_code(fn.__code__, newname)
    new_fn = FunctionType(
        newcode, fn.__globals__, newname, fn.__defaults__, fn.__closure__
    )
    new_fn.__kwdefaults__ = fn.__kwdefaults__
    new_fn.__annotations__ = fn.__annotations__
    return new_fn


class NameConverter(ast.NodeTransformer):
    def __init__(
        self, anal, recurse_sym, call_next_sym, ovld_mangled, code_mangled
    ):
        self.analysis = anal
        self.recurse_sym = recurse_sym
        self.call_next_sym = call_next_sym
        self.ovld_mangled = ovld_mangled
        self.code_mangled = code_mangled

    def visit_Name(self, node):
        if node.id == self.recurse_sym:
            return ast.copy_location(
                old_node=node,
                new_node=ast.Name(self.ovld_mangled, ctx=node.ctx),
            )
        elif node.id == self.call_next_sym:
            raise UsageError("call_next should be called right away")
        else:
            return node

    def visit_Call(self, node):
        if not isinstance(node.func, ast.Name) or node.func.id not in (
            self.recurse_sym,
            self.call_next_sym,
        ):
            return self.generic_visit(node)

        if any(isinstance(arg, ast.Starred) for arg in node.args):
            return self.generic_visit(node)

        cn = node.func.id == self.call_next_sym

        def _make_lookup_call(key, arg):
            value = ast.NamedExpr(
                target=ast.Name(id=f"__TMP{key}", ctx=ast.Store()),
                value=self.visit(arg),
            )
            if self.analysis.lookup_for(key) == "self.map.transform":
                func = ast.Attribute(
                    value=ast.Name(id="__TMPM", ctx=ast.Load()),
                    attr="transform",
                    ctx=ast.Load(),
                )
            else:
                func = ast.Name(id="type", ctx=ast.Load())
            return ast.Call(
                func=func,
                args=[value],
                keywords=[],
            )

        # type index for positional arguments
        type_parts = [
            _make_lookup_call(i, arg) for i, arg in enumerate(node.args)
        ]

        # type index for keyword arguments
        type_parts += [
            ast.Tuple(
                elts=[
                    ast.Constant(value=kw.arg),
                    _make_lookup_call(kw.arg, kw.value),
                ],
                ctx=ast.Load(),
            )
            for kw in node.keywords
        ]

        if cn:
            type_parts.insert(0, ast.Name(id=self.code_mangled, ctx=ast.Load()))
        method = ast.Subscript(
            value=ast.NamedExpr(
                target=ast.Name(id="__TMPM", ctx=ast.Store()),
                value=ast.Attribute(
                    value=ast.Name(id=self.ovld_mangled, ctx=ast.Load()),
                    attr="map",
                    ctx=ast.Load(),
                ),
            ),
            slice=ast.Tuple(
                elts=type_parts,
                ctx=ast.Load(),
            ),
            ctx=ast.Load(),
        )
        if self.analysis.is_method:
            method = ast.Call(
                func=ast.Attribute(
                    value=method,
                    attr="__get__",
                    ctx=ast.Load(),
                ),
                args=[ast.Name(id="self", ctx=ast.Load())],
                keywords=[],
            )

        new_node = ast.Call(
            func=method,
            args=[
                ast.Name(id=f"__TMP{i}", ctx=ast.Load())
                for i, arg in enumerate(node.args)
            ],
            keywords=[
                ast.keyword(
                    arg=kw.arg,
                    value=ast.Name(id=f"__TMP{kw.arg}", ctx=ast.Load()),
                )
                for kw in node.keywords
            ],
        )
        return ast.copy_location(old_node=node, new_node=new_node)


def _search_names(co, values, glb, closure=None):
    if isinstance(co, CodeType):
        if closure is not None:
            for varname, cell in zip(co.co_freevars, closure):
                if any(cell.cell_contents is v for v in values):
                    yield varname
        for name in co.co_names:
            if any(glb.get(name, None) is v for v in values):
                yield name
        else:
            for ct in co.co_consts:
                yield from _search_names(ct, values, glb)


def adapt_function(fn, ovld, newname):
    """Create a copy of the function with a different name."""
    rec_syms = list(
        _search_names(
            fn.__code__, (recurse, ovld), fn.__globals__, fn.__closure__
        )
    )
    cn_syms = list(
        _search_names(fn.__code__, (call_next,), fn.__globals__, fn.__closure__)
    )
    if rec_syms or cn_syms:
        return recode(
            fn, ovld, rec_syms and rec_syms[0], cn_syms and cn_syms[0], newname
        )
    else:
        return rename_function(fn, newname)


def closure_wrap(tree, fname, names):
    wrap = ast.copy_location(
        ast.FunctionDef(
            name="##create_closure",
            args=ast.arguments(
                posonlyargs=[],
                args=[ast.arg(arg=name) for name in names],
                vararg=None,
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=None,
                defaults=[],
            ),
            body=[
                tree,
                ast.Return(ast.Name(id=fname, ctx=ast.Load())),
            ],
            decorator_list=[],
            returns=None,
        ),
        tree,
    )
    ast.fix_missing_locations(wrap)
    return ast.Module(body=[wrap], type_ignores=[])


def recode(fn, ovld, recurse_sym, call_next_sym, newname):
    ovld_mangled = f"___OVLD{ovld.id}"
    code_mangled = f"___CODE{next(_current)}"
    try:
        src = inspect.getsource(fn)
    except OSError:  # pragma: no cover
        raise OSError(
            f"ovld is unable to rewrite {fn} because it cannot read its source code."
            " It may be an issue with __pycache__, so try to either change the source"
            " to force a refresh, or remove __pycache__ altogether. If that does not work,"
            " avoid calling recurse()/call_next()"
        )
    tree = ast.parse(textwrap.dedent(src))
    new = NameConverter(
        anal=ovld.argument_analysis,
        recurse_sym=recurse_sym,
        call_next_sym=call_next_sym,
        ovld_mangled=ovld_mangled,
        code_mangled=code_mangled,
    ).visit(tree)
    new.body[0].decorator_list = []
    if fn.__closure__:
        new = closure_wrap(new.body[0], "irrelevant", fn.__code__.co_freevars)
    ast.fix_missing_locations(new)
    ast.increment_lineno(new, fn.__code__.co_firstlineno - 1)
    res = compile(new, mode="exec", filename=fn.__code__.co_filename)
    if fn.__closure__:
        res = [x for x in res.co_consts if isinstance(x, CodeType)][0]
    (*_, new_code) = [ct for ct in res.co_consts if isinstance(ct, CodeType)]
    new_closure = tuple(
        [
            fn.__closure__[fn.__code__.co_freevars.index(name)]
            for name in new_code.co_freevars
        ]
    )
    new_fn = FunctionType(
        new_code, fn.__globals__, newname, fn.__defaults__, new_closure
    )
    new_fn.__kwdefaults__ = fn.__kwdefaults__
    new_fn.__annotations__ = fn.__annotations__
    new_fn = rename_function(new_fn, newname)
    new_fn.__globals__[ovld_mangled] = ovld
    new_fn.__globals__[code_mangled] = new_fn.__code__
    return new_fn



try:
    from types import UnionType
except ImportError:  # pragma: no cover
    UnionType = None

import sys
from typing import Union

from ovld.mro import Order, subclasscheck, typeorder


class A:
    pass


class B(A):
    pass


def test_subclasscheck():
    assert subclasscheck(B, A)
    assert not subclasscheck(A, B)
    assert subclasscheck(A, A)
    assert subclasscheck(A, object)
    assert subclasscheck(A, Union[A, int])
    assert subclasscheck(int, Union[A, int])


def test_subclasscheck_generic():
    assert subclasscheck(list[int], list[int])
    assert subclasscheck(list[int], list[object])
    assert not subclasscheck(list[object], list[int])


def test_subclasscheck_type():
    assert subclasscheck(type[int], type[object])


def test_subclasscheck_type_union():
    if sys.version_info >= (3, 10):
        assert subclasscheck(type[int | str], type[UnionType])
    assert subclasscheck(type[Union[int, str]], type[Union])
    assert subclasscheck(type[Union], object)
    assert not subclasscheck(object, type[Union])


def test_typeorder_type_union():
    if sys.version_info >= (3, 10):
        assert typeorder(type[int | str], type[UnionType]) is Order.LESS
    assert typeorder(type[Union[int, str]], type[Union]) is Order.LESS
    assert typeorder(object, type[Union]) is Order.MORE



from dataclasses import dataclass
from typing import Protocol, runtime_checkable

from ovld import Dataclass, ovld


@runtime_checkable
class SupportsRead(Protocol):
    def read(self, amount: int) -> bytes: ...


@runtime_checkable
class SupportsWrite(Protocol):
    def write(self, amount: int) -> bytes: ...


@dataclass
class Reader:
    def read(self):
        return 1

    def write(self):
        return 1


@dataclass
class Writer:
    def write(self):
        return 1


@ovld
def f(x: SupportsRead):
    return "yes"


@ovld
def f(x: SupportsWrite):
    return "yes"


# @ovld
# def f(x: Reader):
#     return "R" + f.next(x)


@ovld  # (priority=-1)
def f(x: Dataclass):
    return "DC"


@ovld
def f(x: int):
    return "w/e"


@ovld
def f(x):
    return "no"


# print(f(1))
print(f(Reader()))
# print(f(Writer()))



from typing import TYPE_CHECKING

from .core import (
    Ovld,
    OvldBase,
    OvldCall,
    OvldMC,
    extend_super,
    is_ovld,
    ovld,
)
from .dependent import (
    Dependent,
    DependentType,
    ParametrizedDependentType,
    dependent_check,
)
from .recode import call_next, recurse
from .typemap import (
    MultiTypeMap,
    TypeMap,
)
from .types import (
    Dataclass,
    Deferred,
    Exactly,
    Intersection,
    StrictSubclass,
    class_check,
    parametrized_class_check,
)
from .utils import (
    BOOTSTRAP,
    MISSING,
    Named,
    keyword_decorator,
)
from .version import version as __version__

if TYPE_CHECKING:  # pragma: no cover
    # Pretend that @ovld is @typing.overload.
    # I can't believe this works.
    from typing import overload as ovld


__all__ = [
    "MultiTypeMap",
    "Ovld",
    "OvldBase",
    "OvldCall",
    "OvldMC",
    "TypeMap",
    "extend_super",
    "is_ovld",
    "ovld",
    "Dependent",
    "ParametrizedDependentType",
    "DependentType",
    "dependent_check",
    "BOOTSTRAP",
    "MISSING",
    "Dataclass",
    "Named",
    "Deferred",
    "Exactly",
    "Intersection",
    "StrictSubclass",
    "class_check",
    "parametrized_class_check",
    "keyword_decorator",
    "call_next",
    "recurse",
    "__version__",
]



"""Utilities to overload functions for multiple types."""

import inspect
import itertools
import sys
import textwrap
import typing
from collections import defaultdict
from dataclasses import dataclass, field, replace
from functools import cached_property, partial
from types import GenericAlias

from .recode import (
    Conformer,
    adapt_function,
    generate_dispatch,
    rename_function,
)
from .typemap import MultiTypeMap, is_type_of_type
from .types import normalize_type
from .utils import UsageError, keyword_decorator

_current_id = itertools.count()


def _fresh(t):
    """Returns a new subclass of type t.

    Each Ovld corresponds to its own class, which allows for specialization of
    methods.
    """
    methods = {}
    if not isinstance(getattr(t, "__doc__", None), str):
        methods["__doc__"] = t.__doc__
    return type(t.__name__, (t,), methods)


@keyword_decorator
def _setattrs(fn, **kwargs):
    for k, v in kwargs.items():
        setattr(fn, k, v)
    return fn


@keyword_decorator
def _compile_first(fn, rename=None):
    def first_entry(self, *args, **kwargs):
        self.compile()
        method = getattr(self, fn.__name__)
        assert method is not first_entry
        return method(*args, **kwargs)

    first_entry._replace_by = fn
    first_entry._rename = rename
    return first_entry


def arg0_is_self(fn):
    sgn = inspect.signature(fn)
    params = list(sgn.parameters.values())
    return params and params[0].name == "self"


@dataclass(frozen=True)
class Arginfo:
    position: typing.Optional[int]
    name: typing.Optional[str]
    required: bool
    ann: type

    @cached_property
    def is_complex(self):
        return isinstance(self.ann, GenericAlias)

    @cached_property
    def canonical(self):
        return self.name if self.position is None else self.position


@dataclass(frozen=True)
class Signature:
    types: tuple
    req_pos: int
    max_pos: int
    req_names: frozenset
    vararg: bool
    priority: float
    is_method: bool = False
    arginfo: list[Arginfo] = field(
        default_factory=list, hash=False, compare=False
    )

    @classmethod
    def extract(cls, fn):
        typelist = []
        sig = inspect.signature(fn)
        max_pos = 0
        req_pos = 0
        req_names = set()
        is_method = False

        arginfo = []
        for i, (name, param) in enumerate(sig.parameters.items()):
            if name == "self":
                assert i == 0
                is_method = True
                continue
            pos = nm = None
            ann = normalize_type(param.annotation, fn)
            if param.kind is inspect._POSITIONAL_ONLY:
                pos = i - is_method
                typelist.append(ann)
                req_pos += param.default is inspect._empty
                max_pos += 1
            elif param.kind is inspect._POSITIONAL_OR_KEYWORD:
                pos = i - is_method
                nm = param.name
                typelist.append(ann)
                req_pos += param.default is inspect._empty
                max_pos += 1
            elif param.kind is inspect._KEYWORD_ONLY:
                nm = param.name
                typelist.append((param.name, ann))
                if param.default is inspect._empty:
                    req_names.add(param.name)
            elif param.kind is inspect._VAR_POSITIONAL:
                raise TypeError("ovld does not support *args")
            elif param.kind is inspect._VAR_KEYWORD:
                raise TypeError("ovld does not support **kwargs")
            arginfo.append(
                Arginfo(
                    position=pos,
                    name=nm,
                    required=param.default is inspect._empty,
                    ann=normalize_type(param.annotation, fn),
                )
            )

        return cls(
            types=tuple(typelist),
            req_pos=req_pos,
            max_pos=max_pos,
            req_names=frozenset(req_names),
            vararg=False,
            is_method=is_method,
            priority=None,
            arginfo=arginfo,
        )


def clsstring(cls):
    if cls is object:
        return "*"
    elif isinstance(cls, tuple):
        key, typ = cls
        return f"{key}: {clsstring(typ)}"
    elif is_type_of_type(cls):
        arg = clsstring(cls.__args__[0])
        return f"type[{arg}]"
    elif hasattr(cls, "__origin__"):
        if cls.__origin__ is typing.Union:
            return "|".join(map(clsstring, cls.__args__))
        else:
            return repr(cls)
    elif hasattr(cls, "__name__"):
        return cls.__name__
    else:
        return repr(cls)


def sigstring(types):
    return ", ".join(map(clsstring, types))


class ArgumentAnalyzer:
    def __init__(self):
        self.name_to_positions = defaultdict(set)
        self.position_to_names = defaultdict(set)
        self.counts = defaultdict(lambda: [0, 0])
        self.complex_transforms = set()
        self.total = 0
        self.is_method = None

    def add(self, fn):
        sig = Signature.extract(fn)
        self.complex_transforms.update(
            arg.canonical for arg in sig.arginfo if arg.is_complex
        )
        for arg in sig.arginfo:
            if arg.position is not None:
                self.position_to_names[arg.position].add(arg.name)
            if arg.name is not None:
                self.name_to_positions[arg.name].add(arg.canonical)

            cnt = self.counts[arg.canonical]
            cnt[0] += arg.required
            cnt[1] += 1

        self.total += 1

        if self.is_method is None:
            self.is_method = sig.is_method
        elif self.is_method != sig.is_method:  # pragma: no cover
            raise TypeError(
                "Some, but not all registered methods define `self`. It should be all or none."
            )

    def compile(self):
        for name, pos in self.name_to_positions.items():
            if len(pos) != 1:
                if all(isinstance(p, int) for p in pos):
                    raise TypeError(
                        f"Argument '{name}' is declared in different positions by different methods. The same argument name should always be in the same position unless it is strictly positional."
                    )
                else:
                    raise TypeError(
                        f"Argument '{name}' is declared in a positional and keyword setting by different methods. It should be either."
                    )

        p_to_n = [
            list(names) for _, names in sorted(self.position_to_names.items())
        ]

        positional = list(
            itertools.takewhile(
                lambda names: len(names) == 1 and isinstance(names[0], str),
                reversed(p_to_n),
            )
        )
        positional.reverse()
        strict_positional = p_to_n[: len(p_to_n) - len(positional)]

        assert strict_positional + positional == p_to_n

        strict_positional_required = [
            f"ARG{pos + 1}"
            for pos, _ in enumerate(strict_positional)
            if self.counts[pos][0] == self.total
        ]
        strict_positional_optional = [
            f"ARG{pos + 1}"
            for pos, _ in enumerate(strict_positional)
            if self.counts[pos][0] != self.total
        ]

        positional_required = [
            names[0]
            for pos, names in enumerate(positional)
            if self.counts[pos + len(strict_positional)][0] == self.total
        ]
        positional_optional = [
            names[0]
            for pos, names in enumerate(positional)
            if self.counts[pos + len(strict_positional)][0] != self.total
        ]

        keywords = [
            name
            for _, (name,) in self.name_to_positions.items()
            if not isinstance(name, int)
        ]
        keyword_required = [
            name for name in keywords if self.counts[name][0] == self.total
        ]
        keyword_optional = [
            name for name in keywords if self.counts[name][0] != self.total
        ]

        return (
            strict_positional_required,
            strict_positional_optional,
            positional_required,
            positional_optional,
            keyword_required,
            keyword_optional,
        )

    def lookup_for(self, key):
        return (
            "self.map.transform" if key in self.complex_transforms else "type"
        )


class _Ovld:
    """Overloaded function.

    A function can be added with the `register` method. One of its parameters
    should be annotated with a type, but only one, and every registered
    function should annotate the same parameter.

    Arguments:
        mixins: A list of Ovld instances that contribute functions to this
            Ovld.
        name: Optional name for the Ovld. If not provided, it will be
            gotten automatically from the first registered function or
            dispatch.
        linkback: Whether to keep a pointer in the parent mixins to this
            ovld so that updates can be propagated. (default: False)
        allow_replacement: Allow replacing a method by another with the
            same signature. (default: True)
    """

    def __init__(
        self,
        *,
        mixins=[],
        bootstrap=None,
        name=None,
        linkback=False,
        allow_replacement=True,
    ):
        """Initialize an Ovld."""
        self.id = next(_current_id)
        self._compiled = False
        self.linkback = linkback
        self.children = []
        self.allow_replacement = allow_replacement
        self.bootstrap = bootstrap
        self.name = name
        self.shortname = name or f"__OVLD{self.id}"
        self.__name__ = name
        self._defns = {}
        self._locked = False
        self.mixins = []
        self.add_mixins(*mixins)
        self.ocls = _fresh(OvldCall)

    @property
    def defns(self):
        defns = {}
        for mixin in self.mixins:
            defns.update(mixin.defns)
        defns.update(self._defns)
        return defns

    @property
    def __doc__(self):
        if not self._compiled:
            self.compile()

        docs = [fn.__doc__ for fn in self.defns.values() if fn.__doc__]
        if len(docs) == 1:
            maindoc = docs[0]
        else:
            maindoc = f"Ovld with {len(self.defns)} methods."

        doc = f"{maindoc}\n\n"
        for fn in self.defns.values():
            fndef = inspect.signature(fn)
            fdoc = fn.__doc__
            if not fdoc or fdoc == maindoc:
                doc += f"{self.__name__}{fndef}\n\n"
            else:
                if not fdoc.strip(" ").endswith("\n"):
                    fdoc += "\n"
                fdoc = textwrap.indent(fdoc, " " * 4)
                doc += f"{self.__name__}{fndef}\n{fdoc}\n"
        return doc

    @property
    def __signature__(self):
        if not self._compiled:
            self.compile()

        sig = inspect.signature(self._dispatch)
        if not self.argument_analysis.is_method:
            sig = inspect.Signature(
                [v for k, v in sig.parameters.items() if k != "self"]
            )
        return sig

    def lock(self):
        self._locked = True

    def _attempt_modify(self):
        if self._locked:
            raise Exception(f"ovld {self} is locked for modifications")

    def add_mixins(self, *mixins):
        self._attempt_modify()
        for mixin in mixins:
            if self.linkback:
                mixin.children.append(self)
            if mixin._defns and self.bootstrap is None:
                self.bootstrap = mixin.bootstrap
        self.mixins += mixins

    def _key_error(self, key, possibilities=None):
        typenames = sigstring(key)
        if not possibilities:
            return TypeError(
                f"No method in {self} for argument types [{typenames}]"
            )
        else:
            hlp = ""
            for p, prio, spc in possibilities:
                hlp += f"* {p.__name__}  (priority: {prio}, specificity: {list(spc)})\n"
            return TypeError(
                f"Ambiguous resolution in {self} for"
                f" argument types [{typenames}]\n"
                f"Candidates are:\n{hlp}"
                "Note: you can use @ovld(priority=X) to give higher priority to an overload."
            )

    def rename(self, name):
        """Rename this Ovld."""
        self.name = name
        self.__name__ = name

    def _set_attrs_from(self, fn):
        """Inherit relevant attributes from the function."""
        if self.bootstrap is None:
            self.bootstrap = arg0_is_self(fn)

        if self.name is None:
            self.name = f"{fn.__module__}.{fn.__qualname__}"
            self.shortname = fn.__name__
            self.__name__ = fn.__name__
            self.__qualname__ = fn.__qualname__
            self.__module__ = fn.__module__

    def _maybe_rename(self, fn):
        if hasattr(fn, "rename"):
            return rename_function(fn, f"{self.__name__}.{fn.rename}")
        else:
            return fn

    def compile(self):
        """Finalize this overload.

        This will populate the type maps and replace the functions decorated
        with _compile_first (__call__, __get__, etc.) with versions that assume
        the ovld has been compiled.

        This will also lock this ovld's parent mixins to prevent their
        modification.
        """
        for mixin in self.mixins:
            if self not in mixin.children:
                mixin.lock()

        cls = type(self)
        if self.name is None:
            self.name = self.__name__ = f"ovld{self.id}"

        name = self.__name__
        self.map = MultiTypeMap(name=name, key_error=self._key_error)

        # Replace the appropriate functions by their final behavior
        for method in dir(cls):
            value = getattr(cls, method)
            repl = getattr(value, "_replace_by", None)
            if repl:
                repl = self._maybe_rename(repl)
                setattr(cls, method, repl)

        target = self.ocls if self.bootstrap else cls

        anal = ArgumentAnalyzer()
        for key, fn in list(self.defns.items()):
            anal.add(fn)
        self.argument_analysis = anal
        dispatch = generate_dispatch(anal)
        self._dispatch = dispatch
        target.__call__ = rename_function(dispatch, f"{name}.dispatch")

        for key, fn in list(self.defns.items()):
            self.register_signature(key, fn)

        self._compiled = True

    @_compile_first
    def resolve(self, *args):
        """Find the correct method to call for the given arguments."""
        return self.map[tuple(map(self.map.transform, args))]

    def register_signature(self, sig, orig_fn):
        """Register a function for the given signature."""
        fn = adapt_function(
            orig_fn, self, f"{self.__name__}[{sigstring(sig.types)}]"
        )
        # We just need to keep the Conformer pointer alive for jurigged
        # to find it, if jurigged is used with ovld
        fn._conformer = Conformer(self, orig_fn, fn)
        self.map.register(sig, fn)
        return self

    def register(self, fn=None, priority=0):
        """Register a function."""
        if fn is None:
            return partial(self._register, priority=priority)
        return self._register(fn, priority)

    def _register(self, fn, priority):
        """Register a function."""

        self._attempt_modify()

        self._set_attrs_from(fn)

        sig = replace(Signature.extract(fn), priority=priority)
        if not self.allow_replacement and sig in self._defns:
            raise TypeError(
                f"There is already a method for {sigstring(sig.types)}"
            )
        self._defns[sig] = fn

        self._update()
        return self

    def unregister(self, fn):
        """Unregister a function."""
        self._attempt_modify()
        self._defns = {sig: f for sig, f in self._defns.items() if f is not fn}
        self._update()

    def _update(self):
        if self._compiled:
            self.compile()
        for child in self.children:
            child._update()

    def copy(self, mixins=[], linkback=False):
        """Create a copy of this Ovld.

        New functions can be registered to the copy without affecting the
        original.
        """
        return _fresh(_Ovld)(
            bootstrap=self.bootstrap,
            mixins=[self, *mixins],
            linkback=linkback,
        )

    def variant(self, fn=None, priority=0, **kwargs):
        """Decorator to create a variant of this Ovld.

        New functions can be registered to the variant without affecting the
        original.
        """
        ov = self.copy(**kwargs)
        if fn is None:
            return partial(ov.register, priority=priority)
        else:
            ov.register(fn, priority=priority)
            return ov

    @_compile_first
    def __get__(self, obj, cls):
        if obj is None:
            return self
        key = self.shortname
        rval = obj.__dict__.get(key, None)
        if rval is None:
            obj.__dict__[key] = rval = self.ocls(self, obj)
        return rval

    @_compile_first
    def __getitem__(self, t):
        if not isinstance(t, tuple):
            t = (t,)
        return self.map[t]

    @_compile_first
    @_setattrs(rename="dispatch")
    def __call__(self, *args):  # pragma: no cover
        """Call the overloaded function.

        This should be replaced by an auto-generated function.
        """
        key = tuple(map(self.map.transform, args))
        method = self.map[key]
        return method(*args)

    @_compile_first
    @_setattrs(rename="next")
    def next(self, *args):
        """Call the next matching method after the caller, in terms of priority or specificity."""
        fr = sys._getframe(1)
        key = (fr.f_code, *map(self.map.transform, args))
        method = self.map[key]
        return method(*args)

    def __repr__(self):
        return f"<Ovld {self.name or hex(id(self))}>"

    @_compile_first
    def display_methods(self):
        self.map.display_methods()

    @_compile_first
    def display_resolution(self, *args, **kwargs):
        self.map.display_resolution(*args, **kwargs)


def is_ovld(x):
    """Return whether the argument is an ovld function/method."""
    return isinstance(x, _Ovld)


class OvldCall:
    """Context for an Ovld call."""

    def __init__(self, ovld, bind_to):
        """Initialize an OvldCall."""
        self.ovld = ovld
        self.map = ovld.map
        self.obj = bind_to

    @property
    def __name__(self):
        return self.ovld.__name__

    @property
    def __doc__(self):
        return self.ovld.__doc__

    @property
    def __signature__(self):
        return self.ovld.__signature__

    def next(self, *args):
        """Call the next matching method after the caller, in terms of priority or specificity."""
        fr = sys._getframe(1)
        key = (fr.f_code, *map(self.map.transform, args))
        method = self.map[key]
        return method(self.obj, *args)

    def resolve(self, *args):
        """Find the right method to call for the given arguments."""
        return self.map[tuple(map(self.map.transform, args))].__get__(self.obj)

    def __call__(self, *args):  # pragma: no cover
        """Call this overloaded function.

        This should be replaced by an auto-generated function.
        """
        key = tuple(map(self.map.transform, args))
        method = self.map[key]
        return method(self.obj, *args)


def Ovld(*args, **kwargs):
    """Returns an instance of a fresh Ovld class."""
    return _fresh(_Ovld)(*args, **kwargs)


def extend_super(fn):
    """Declare that this method extends the super method with more types.

    This produces an ovld using the superclass method of the same name,
    plus this definition and others with the same name.
    """
    if not is_ovld(fn):
        fn = ovld(fn, fresh=True)
    fn._extend_super = True
    return fn


class ovld_cls_dict(dict):
    """A dict for use with OvldMC.__prepare__.

    Setting a key that already corresponds to an Olvd extends that Ovld.
    """

    def __init__(self, bases):
        self._bases = bases

    def __setitem__(self, attr, value):
        prev = None
        if attr in self:
            prev = self[attr]
            if inspect.isfunction(prev):
                prev = ovld(prev, fresh=True)
            elif not is_ovld(prev):  # pragma: no cover
                prev = None
        elif is_ovld(value) and getattr(value, "_extend_super", False):
            mixins = []
            for base in self._bases:
                if (candidate := getattr(base, attr, None)) is not None:
                    if is_ovld(candidate) or inspect.isfunction(candidate):
                        mixins.append(candidate)
            if mixins:
                prev, *others = mixins
                if is_ovld(prev):
                    prev = prev.copy()
                else:
                    prev = ovld(prev, fresh=True)
                for other in others:
                    if is_ovld(other):
                        prev.add_mixins(other)
                    else:
                        prev.register(other)
        else:
            prev = None

        if prev is not None:
            if is_ovld(value) and prev is not value:
                if prev.name is None:
                    prev.rename(value.name)
                prev.add_mixins(value)
                value = prev
            elif inspect.isfunction(value):
                prev.register(value)
                value = prev

        super().__setitem__(attr, value)


class OvldMC(type):
    """Metaclass that allows overloading.

    A class which uses this metaclass can define multiple functions with
    the same name and different type signatures.
    """

    def create_subclass(cls, *bases, name=None):
        """Create a new subclass with the given extra bases."""
        name = name or "UnnamedOvld"
        bases = (cls, *bases)
        return type(cls)(name, bases, cls.__prepare__(name, bases))

    @classmethod
    def __prepare__(cls, name, bases):
        d = ovld_cls_dict(bases)

        names = set()
        for base in bases:
            names.update(dir(base))

        for name in names:
            values = [getattr(base, name, None) for base in bases]
            ovlds = [v for v in values if is_ovld(v)]
            mixins = [
                v for v in ovlds[1:] if getattr(v, "_extend_super", False)
            ]
            if mixins:
                o = ovlds[0].copy(mixins=mixins)
                others = [v for v in values if v is not None and not is_ovld(v)]
                for other in others:
                    o.register(other)
                o.rename(name)
                d[name] = o

        return d


class OvldBase(metaclass=OvldMC):
    """Base class that allows overloading of methods."""


def _find_overload(fn, **kwargs):
    fr = sys._getframe(1)  # We typically expect to get to frame 3.
    while fr and fn.__code__ not in fr.f_code.co_consts:
        # We are basically searching for the function's code object in the stack.
        # When a class/function A is nested in a class/function B, the former's
        # code object is in the latter's co_consts. If ovld is used as a decorator,
        # on A, then necessarily we are inside the execution of B, so B should be
        # on the stack and we should be able to find A's code object in there.
        fr = fr.f_back

    if not fr:
        raise UsageError("@ovld only works as a decorator.")

    dispatch = fr.f_locals.get(fn.__name__, None)

    if dispatch is None:
        dispatch = _fresh(_Ovld)(**kwargs)
    elif not is_ovld(dispatch):  # pragma: no cover
        raise TypeError("@ovld requires Ovld instance")
    elif kwargs:  # pragma: no cover
        raise TypeError("Cannot configure an overload that already exists")
    return dispatch


@keyword_decorator
def ovld(fn, priority=0, fresh=False, **kwargs):
    """Overload a function.

    Overloading is based on the function name.

    The decorated function should have one parameter annotated with a type.
    Any parameter can be annotated, but only one, and every overloading of a
    function should annotate the same parameter.

    The decorator optionally takes keyword arguments, *only* on the first
    use.

    Arguments:
        fn: The function to register.
        priority: The priority of the function in the resolution order.
        fresh: Whether to create a new ovld or try to reuse an existing one.
        mixins: A list of Ovld instances that contribute functions to this
            Ovld.
        name: Optional name for the Ovld. If not provided, it will be
            gotten automatically from the first registered function or
            dispatch.
        linkback: Whether to keep a pointer in the parent mixins to this
            ovld so that updates can be propagated. (default: False)
    """
    if fresh:
        dispatch = _fresh(_Ovld)(**kwargs)
    else:
        dispatch = _find_overload(fn, **kwargs)
    return dispatch.register(fn, priority=priority)


__all__ = [
    "Ovld",
    "OvldBase",
    "OvldCall",
    "OvldMC",
    "extend_super",
    "is_ovld",
    "ovld",
]



from multimethod import multimethod as dispatch
from plum import dispatch
from runtype import multidispatch as dispatch

from ovld import ovld as dispatch


@dispatch
def f(x: type[int]):
    return "ah"


@dispatch
def f(x: type[object]):
    return "something else"


@dispatch
def f(x: type[dict[str, object]]):
    return ("didi", x)


print(f(int))
print(f(str))
print(f(dict[str, int]))
print(f(dict[int, int]))



import typing
from dataclasses import dataclass
from enum import Enum
from graphlib import TopologicalSorter


class Order(Enum):
    LESS = -1
    MORE = 1
    SAME = 0
    NONE = None

    def opposite(self):
        if self is Order.LESS:
            return Order.MORE
        elif self is Order.MORE:
            return Order.LESS
        else:
            return self


@dataclass
class TypeRelationship:
    order: Order
    matches: bool = None


def _issubclass(t1, t2):
    try:
        return issubclass(t1, t2)
    except TypeError:
        try:
            return isinstance(t1, t2)
        except TypeError:  # pragma: no cover
            return False


def typeorder(t1, t2):
    """Order relation between two types.

    Returns a member of the Order enum.

    * typeorder(t1, t2) is Order.LESS   if t2 is more general than t1
    * typeorder(t1, t2) is Order.SAME   if t1 is equal to t2
    * typeorder(t1, t2) is Order.MORE   if t1 is more general than t2
    * typeorder(t1, t2) is Order.NONE   if they cannot be compared
    """
    if t1 == t2:
        return Order.SAME

    t1 = getattr(t1, "__proxy_for__", t1)
    t2 = getattr(t2, "__proxy_for__", t2)

    if (
        hasattr(t1, "__typeorder__")
        and (result := t1.__typeorder__(t2)) is not NotImplemented
    ):
        return result.order
    elif (
        hasattr(t2, "__typeorder__")
        and (result := t2.__typeorder__(t1)) is not NotImplemented
    ):
        return result.order.opposite()

    o1 = getattr(t1, "__origin__", None)
    o2 = getattr(t2, "__origin__", None)

    if o2 is typing.Union:
        if t1 is typing.Union:
            return Order.MORE
        compare = [
            x for t in t2.__args__ if (x := typeorder(t1, t)) is not Order.NONE
        ]
        if not compare:
            return Order.NONE
        elif any(x is Order.LESS or x is Order.SAME for x in compare):
            return Order.LESS
        else:
            return Order.MORE

    if o1 is typing.Union:
        return typeorder(t2, t1).opposite()

    if o2 and not o1:
        return typeorder(t2, t1).opposite()

    if o1:
        if not o2:  # or getattr(t2, "__args__", None) is None:
            order = typeorder(o1, t2)
            if order is order.SAME:
                order = order.LESS
            return order

        if (order := typeorder(o1, o2)) is not Order.SAME:
            return order

        args1 = getattr(t1, "__args__", ())
        args2 = getattr(t2, "__args__", ())

        if args1 and not args2:
            return Order.LESS
        if args2 and not args1:
            return Order.MORE
        if len(args1) != len(args2):
            return Order.NONE

        ords = [typeorder(a1, a2) for a1, a2 in zip(args1, args2)]
        if Order.MORE in ords and Order.LESS in ords:
            return Order.NONE
        elif Order.NONE in ords:
            return Order.NONE
        elif Order.MORE in ords:
            return Order.MORE
        elif Order.LESS in ords:
            return Order.LESS
        else:  # pragma: no cover
            # Not sure when t1 != t2 and that happens
            return Order.SAME

    sx = _issubclass(t1, t2)
    sy = _issubclass(t2, t1)
    if sx and sy:  # pragma: no cover
        # Not sure when t1 != t2 and that happens
        return Order.SAME
    elif sx:
        return Order.LESS
    elif sy:
        return Order.MORE
    else:
        return Order.NONE


def subclasscheck(t1, t2):
    """Check whether t1 is a "subclass" of t2."""
    if t1 == t2:
        return True

    t1 = getattr(t1, "__proxy_for__", t1)
    t2 = getattr(t2, "__proxy_for__", t2)

    if (
        hasattr(t2, "__typeorder__")
        and (result := t2.__typeorder__(t1)) is not NotImplemented
    ):
        return result.matches

    o1 = getattr(t1, "__origin__", None)
    o2 = getattr(t2, "__origin__", None)

    if o2 is typing.Union:
        return t1 is typing.Union or any(
            subclasscheck(t1, t) for t in t2.__args__
        )
    elif o1 is typing.Union:
        return t2 is typing.Union or all(
            subclasscheck(t, t2) for t in t1.__args__
        )

    if not isinstance(o1, type):
        o1 = None
    if not isinstance(o2, type):
        o2 = None

    if o1 or o2:
        o1 = o1 or t1
        o2 = o2 or t2
        if _issubclass(o1, o2):
            if o2 is t2:  # pragma: no cover
                return True
            else:
                args1 = getattr(t1, "__args__", ())
                args2 = getattr(t2, "__args__", ())
                if len(args1) != len(args2):
                    return False
                return all(
                    subclasscheck(a1, a2) for a1, a2 in zip(args1, args2)
                )
        else:
            return False
    else:
        return _issubclass(t1, t2)


def sort_types(cls, avail):
    # We filter everything except subclasses and dependent types that *might* cover
    # the object represented by cls.
    avail = [t for t in avail if subclasscheck(cls, t)]
    deps = {t: set() for t in avail}
    for i, t1 in enumerate(avail):
        for t2 in avail[i + 1 :]:
            # NOTE: this is going to scale poorly when there's a hundred Literal in the pool
            order = typeorder(t1, t2)
            if order is Order.LESS:
                deps[t2].add(t1)
            elif order is Order.MORE:
                deps[t1].add(t2)
    sorter = TopologicalSorter(deps)
    sorter.prepare()
    while sorter.is_active():
        nodes = sorter.get_ready()
        yield nodes
        for n in nodes:
            sorter.done(n)



import inspect
from collections import defaultdict


class ArgumentAnalyzer:
    def __init__(self):
        self.name_to_positions = defaultdict(set)
        self.position_to_names = defaultdict(set)
        self.counts = defaultdict(lambda: [0, 0])
        self.total = 0
        self.is_method = None

    def add(self, fn):
        sig = inspect.signature(fn)
        is_method = False
        for i, (name, param) in enumerate(sig.parameters.items()):
            if name == "self":
                assert i == 0
                is_method = True
                continue
            if param.kind is inspect._POSITIONAL_ONLY:
                cnt = self.counts[i]
                self.position_to_names[i].add(None)
            elif param.kind is inspect._POSITIONAL_OR_KEYWORD:
                cnt = self.counts[i]
                self.position_to_names[i].add(param.name)
                self.name_to_positions[param.name].add(i)
            elif param.kind is inspect._KEYWORD_ONLY:
                cnt = self.counts[param.name]
                self.name_to_positions[param.name].add(param.name)
            elif param.kind is inspect._VAR_POSITIONAL:
                raise TypeError("ovld does not support *args")
            elif param.kind is inspect._VAR_KEYWORD:
                raise TypeError("ovld does not support **kwargs")

            cnt[0] += 1 if param.default is inspect._empty else 0
            cnt[1] += 1

        self.total += 1

        if self.is_method is None:
            self.is_method = is_method
        elif self.is_method != is_method:
            raise TypeError(
                "Some, but not all registered methods define `self`. It should be all or none."
            )

    def compile(self):
        if any(
            len(pos) != 1
            for _name, pos in self.name_to_positions.items()
            if (name := _name) is not None
        ):
            raise TypeError(
                f"Argument {name} is found both in a positional and keyword setting."
            )
        npositional = 0
        positional = []
        for pos, names in sorted(self.position_to_names.items()):
            required, total = self.counts[pos]
            name = f"_ovld_arg{pos}"
            if len(names) == 1 and total == self.total:
                name = list(names)[0]
            else:
                npositional = pos + 1
            positional.append((name, required == self.total))

        keywords = []
        for key, (name,) in self.name_to_positions.items():
            if isinstance(name, int):
                pass  # ignore positional arguments
            else:
                assert key == name
                required, total = self.counts[key]
                keywords.append((name, required == self.total))

        return positional[:npositional], positional[npositional:], keywords


anal = ArgumentAnalyzer()


def foo(fn):
    anal.add(fn)


@foo
def f(x, y, *, zou):
    pass


@foo
def f(x, z, *, fax, coop):
    pass


def main():
    po, pn, kw = anal.compile()

    print(po)
    print(pn)
    print(kw)

    template = """
    def __DISPATCH__(self, {args}):
        KWARGS = {{}}
        {body}
        method = self.map[{lookup}]
        return method({posargs}, **KWARGS)
    """

    args = []
    body = [""]
    posargs = []
    lookup = []

    for k, necessary in po:
        txt = k
        if not necessary:
            txt += "=MISSING"
        args.append(txt)

    print(
        template.format(
            args=", ".join(args),
            posargs=", ".join(posargs),
            body="\n    ".join(body),
            lookup=", ".join(lookup),
        )
    )


if __name__ == "__main__":
    main()



from ovld.core import ovld

# @ovld(priority=1000)
# def f(x: object):
#     return "A"


# @ovld
# def f(x: Number):
#     return "B"


# @ovld
# def f(x: Dependent[int, lambda x: x < 0]):
#     return "C"


# @ovld
# def f(x: int):
#     return "D"


# @ovld
# def f(x: str):
#     return "E"


# @ovld(priority=-1)
# def f(x: object):
#     return "F"


# f.display_resolution(123)
# print("=" * 50)
# f.display_resolution("hello")


@ovld
def add(x: list, y: list):
    return [add(a, b) for a, b in zip(x, y)]


@ovld
def add(x: tuple, y: tuple):
    return tuple(add(a, b) for a, b in zip(x, y))


@ovld
def add(x: dict, y: dict):
    return {k: add(v, y[k]) for k, v in x.items()}


@ovld
def add(x: object, y: object):
    return x + y


add([[[1]]], [[[[2]]]])
